#  C O N S T R U C T I V I S T__ __
#   _____ ____   ____   /  |/  // /
#  / ___// __ \ / __ \ / /|_/ // /
# / /__ / /_/ // / / // /  / // /___
# \___/ \____//_/ /_//_/  /_//_____/
#  M A C H I N E   L E A R N I N G
#
#
# A Project by
# Thomas Schmid | UNIVERSITÄT LEIPZIG
# www.constructivist.ml
#
# Code Author: Dmitrij Denisenko
# Licence: MIT


"""Define the Z-Sigma deconstruction.

"""


from conML.domain.deconstruction.abstract import Abstract
from conML.shared.logger import DeconstructionInfo


__all__ = (
    "ZSigmaDeconstructor",
)


class ZSigmaDeconstructor(Abstract):
    """

    Notes:
        Please see help(Abstract) for more info.

    """
    RELATION = ("Sigma", "Z")
    MIN_N_FEATURE = 2

    def __init__(self, *args):
        super().__init__(*args)

    def deconstruct(self, relative, log):
        """Run the ZSigma deconstruction.

        Args:
            relative (PragmaticMachineLearningModel): Relative of the currently
                deconstructed model.
            log (DeconstructionInfo): Container holding information
                of the deconstruction process.

        Returns:
            DeconstructionInfo:
                Logging information.

        """
        if self.time_constraint(relative):
            overlapping = self.identify_overlapping_features()
            if self.column_constraint(overlapping):
                meta_model = self.model_fusion(relative)

                if self.settings.force_time_expansion:
                    return self.reconstructor.get_pml(meta_model,
                                                      overlapping,
                                                      self.pragmatic.model)
                else:
                    return self.reconstructing(overlapping,
                                               meta_model,
                                               relative,
                                               log)

        return self.deconstruction_failed(relative, log)

    def time_constraint(self, relative):
        """Check if the relative is satisfying the time constraint.

        Args:
            relative (PragmaticMachineLearningModel): Relative model.

        Returns:
            bool:
                True, if the time contraint is fullfilled.

        """
        if self.settings.deconst_max_distance_t == 0.0:
            return self.overlapping_time(self.pragmatic, relative)

        elif 0.0 < self.settings.deconst_max_distance_t < 1.0:
            return self.enclosing_time(self.pragmatic, relative)

        elif self.settings.deconst_max_distance_t == 1.0:
            return True

    def overlapping_time(self, prag_model, relative_model):
        """Check if the time values between two pragmtic machine learning
        models overlap.

        Args:
            prag_model (PragmaticMachineLearningModel): Model to deconstruct.
            relative_model (PragmaticMachineLearningModel): Relative.

        Returns:
            bool:
                True, if time stamps overlapping else False.

        """
        return (relative_model.min_timestamp >= prag_model.max_timestamp or
                prag_model.min_timestamp >= relative_model.max_timestamp)

    def enclosing_time(self, prag_model, relative_model):
        """Check if the time values of the deconstructing model enclose
        the relative model.


        Args:
            prag_model (PragmaticMachineLearningModel): Model to deconstruct.
            relative_model (PragmaticMachineLearningModel): Relative.

        Returns:
            bool:
                True, if time stamps of deconstructed model enclose
                the relative time stamps.

        """
        m_dash_max_time = max(prag_model.max_timestamp,
                              relative_model.max_timestamp)
        m_dash_min_time = min(prag_model.min_timestamp,
                              relative_model.min_timestamp)
        relative_m_dash_time = self.settings.deconst_max_distance_t*(
                m_dash_max_time-m_dash_min_time)

        return (
            (relative_model.min_timestamp - prag_model.max_timestamp) <
            relative_m_dash_time or
            (prag_model.min_timestamp - relative_model.max_timestamp) <
            relative_m_dash_time
        )

    def identify_overlapping_features(self):
        """Build learnblock on basis of the feature intersection of two other
        learnblocks.

        Returns:
            PandasBlock:
                Learnblock

        """
        return self.block.same_features_fusion(self.relative_block)

    def column_constraint(self, overlapped_block):
        """Check if the new learnblock has enough features.

        Args:
            overlapped_block (PandasBlock): Learnblock.

        Returns:
            bool:
                Learnblock with more than 2 features.

        """
        return overlapped_block.n_columns(effective=True) >= self.MIN_N_FEATURE

    def reconstructing(self, overlapped_block, meta_model, relative, log):
        """Reconstruct the new combined pragmatic machine learning
        model.

        Args:
            overlapped_block (PandasBlock): Learnblock with more than 2
                Features.
            meta_model (MetaData): Meta data of combined pragmatic machine
                learning.
            relative (PragmaticMachineLearningModel): Relative.
            log (DeconstructionInfo): Logging information.

        Returns:
            DeconstructionInfo:
                Holds information about T Sigma deconstruction.

        """
        _, pragmatics, _ = self.reconstructor.reconstruct(
            self.level, overlapped_block, which_models=None,
            meta=meta_model, all_m=True)

        if pragmatics:
            winners = {m.accuracy: m for m in pragmatics}
            winner = winners[max(winners.keys())]
            deleted = self.knowledge.remove_dependent(relative)
            self.knowledge.insert(winner)
            log.reconstructed_info.append(winner)
            log.inserted.append(winner)
            log.deleted.extend(deleted)
            log.state = DeconstructionInfo.State.DECONSTRUCTED
            return log

        else:
            return self.deconstruction_failed(relative, log)

    def deconstruction_failed(self, relative, log):
        """Fill the DeconstrucionInfo object if the T Sigma deconstruction
        failed.

        Args:
            relative (PragmaticMachineLearningModel): Relative.
            log (DeconstructionInfo): Logging info.

        Returns:
            DeconstructionInfo

        """
        modification = self.model_disposal(
            self.pragmatic,
            relative,
        )
        log.inserted.extend(modification.inserted)
        log.deleted.extend(modification.deleted)
        log.state = DeconstructionInfo.State.FAILED
        return log



#  C O N S T R U C T I V I S T__ __
#   _____ ____   ____   /  |/  // /
#  / ___// __ \ / __ \ / /|_/ // /
# / /__ / /_/ // / / // /  / // /___
# \___/ \____//_/ /_//_/  /_//_____/
#  M A C H I N E   L E A R N I N G
#
#
# A Project by
# Thomas Schmid | UNIVERSITÄT LEIPZIG
# www.constructivist.ml
#
# Code Author: Dmitrij Denisenko
# Licence: MIT


"""Define application specific exceptions.

"""


class BaseError(Exception):
    """Base class for all application specific exceptions.

    """
    pass


class ModeError(BaseError):
    """Will be thrown, if the modes for Constructor and Reconstructor
    are not identical.

    """
    pass


class IndexOutOfBucket(BaseError):
    """Will be thrown, if the given index for the requested row does not
    correspond to any row within the stored block.

    """
    pass


class ConfigurationFileError(BaseError):
    """Will be thrown, if the configuration file cannot be read.

    """
    pass



#  C O N S T R U C T I V I S T__ __
#   _____ ____   ____   /  |/  // /
#  / ___// __ \ / __ \ / /|_/ // /
# / /__ / /_/ // / / // /  / // /___
# \___/ \____//_/ /_//_/  /_//_____/
#  M A C H I N E   L E A R N I N G
#
#
# A Project by
# Thomas Schmid | UNIVERSITÄT LEIPZIG
# www.constructivist.ml
#
# Code Author: Dmitrij Denisenko
# Licence: MIT


"""Define the base class for all deconstuction types.

"""


from collections import namedtuple

from conML.shared.logger import DeconstructionInfo


class Abstract:
    """Baseclass for all other deconstruction classes.

    Args:
        level (int):
            Level specifiction on which the deconstruction is performed.
        knowledge (KnowledgeDatabase):
            Deconstruction changes writting to this database.
        settings (DeconstructionSettings):
            Controlls the behavior of the deconstruction.
        reconstructor (Reconstructor):
            Performs the reconstruction step within the deconstruction.
        pragmatic (PragmaticMachineLearningModel):
            Model which should be deconstructed.
        relatives (list): List of all relatives of pragmatic.
        result_queue (deque):
            Used as a container for all generated DeconstructionInfo.

    """
    def __init__(self, level, knowledge, settings, reconstructor, pragmatic,
                 relatives, result_queue):
        self.level = level
        self.knowledge = knowledge
        self.settings = settings
        self.reconstructor = reconstructor
        self.pragmatic = pragmatic
        self.relatives = relatives
        self.result_queue = result_queue
        self.block = pragmatic.trained_with(knowledge)

        self.relative_block = None

    def run(self):
        """Start the deconstruction.

        """
        for relative in self.relatives:
            if relative in self.knowledge:
                self.relative_block = relative.trained_with(self.knowledge)

                log = self.log_info(relative, self.pragmatic)
                self.deconstruct(relative, log)
                self.result_queue.append(log)

                if (log.state == DeconstructionInfo.State.DECONSTRUCTED
                        and self.settings.deconst_mode == "minimal"):
                    return

    def deconstruct(self, *args, **kwargs):
        raise NotImplementedError

    def log_info(self, relative, pragmatic):
        """Build a container for holding important information during
        deconstruction.

        Args:
            relative (PragmaticMachineLearningModel):
                Relative of the model, which should be deconstructed.
            pragmatic (PragmaticMachineLearningModel):
                Model, which should be deconstructed.

        Returns:
            DeconstructionInfo

        """
        try:
            relationship = self.RELATION
        except AttributeError:
            relationship = None

        return DeconstructionInfo(
            relative=relative,
            pragmatic=pragmatic,
            relationship=relationship
        )

    def push_ids_to_reconstructor(self, n_ids):
        """Push valid identifier to the reconstructor.

        Args:
            n_ids (int): Amount of how many valid identifier should be
                generated.

        """
        for _ in range(n_ids):
            identifier = self.knowledge.get_next_identifier(self.level)
            self.reconstructor.identifier_queue.put(identifier)

    def model_disposal(self, pragmatic, relative):
        """Remove models depending on the chosen deconstructon strategy.

        Args:
            pragmatic (PragmaticMachineLearningModel):
                Model, which should be deconstructed.
            relative (PragmaticMachineLearningModel):
                Relative of the model, which should be deconstructed.

        Returns:
            namedtuple:
                Contained which models are deleted and inserted.

        """
        Disposal = namedtuple("Disposal", "inserted, deleted")
        if self.settings.deconst_strategy == "conservative":
            if pragmatic != self.pragmatic:
                deleted = self.knowledge.remove_dependent(pragmatic)
            else:
                deleted = [pragmatic]

            return Disposal(deleted=deleted, inserted=[])

        elif self.settings.deconst_strategy == "integrative":
            self.knowledge.insert(pragmatic)
            return Disposal(inserted=[pragmatic], deleted=[])

        elif self.settings.deconst_strategy == "oppurtunistic":
            if self.block.n_rows() > self.relative_block.n_rows():
                deleted = self.knowledge.remove_dependent(relative)
                self.knowledge.insert(pragmatic)
                return Disposal(inserted=[pragmatic], deleted=deleted)

            else:
                if pragmatic != self.pragmatic:
                    deleted = self.knowledge.remove_dependent(pragmatic)
                else:
                    deleted = [pragmatic]
                self.knowledge.insert(relative)
                return Disposal(inserted=[relative], deleted=deleted)

    def calculate_reliability(self, predicts_a, predicts_b):
        """Calculate the reliability between two voters.

        Args:
            predicts_a (Iterable): Containing the first votes.
            predicts_b (Iterable): Containing the second votes.

        Returns:
            float:
                Calculated reliability value, between -1 and 1.

        """
        predicts_b = [int(i) for i in predicts_b]
        predicts_a = [int(i) for i in predicts_a]
        predictions = [predicts_b, predicts_a]
        return self.reconstructor.krippendorff_alpha(predictions)

    def model_fusion(self, relative):
        """Combine the PragmaticMachineLearningModel, which is currently
        in deconstruction with one of this relatives.

        Args:
            relative (PragmaticMachineLearningModel):
                Relative of the model, which should be deconstructed.

        Returns:
            MetaData:
                Combined meta data of two PragmaticMachineLearningModels.

        """
        self.push_ids_to_reconstructor(n_ids=1)
        return self.pragmatic.fusion(
            relative,
            self.level,
            self.reconstructor.identifier_queue.get()
        )



"""Defines the feature selection process.

"""


from conML.shared.logger import SelectionInfo, WinnerSelectionInfo


class WinnerSelector:
    def determine_winner(self, reliability_dict, ml_models):
        """Determine the winner among all pragmatic machine learning models.

        Args:
            reliability_dict (dict): Map reliability to list of models.
            ml_models (list): Supervised machine learnign models.

        Returns:
            tuple:
                Logging Info and PragmaticMachineLearningModel.

        """
        if not reliability_dict:
            return (
                WinnerSelectionInfo(WinnerSelectionInfo.State.FAILED),
                None
            )

        reliability_winner = self.models_with_the_biggest_reliability(reliability_dict)
        domain_winner = self.models_with_the_smallest_domain(reliability_winner)
        accuracy_winner = self.models_with_the_biggest_accuracy(domain_winner)
        priority_winner = self.model_with_highest_priority_winner(
            accuracy_winner, ml_models
        )

        return (
            WinnerSelectionInfo(
                WinnerSelectionInfo.State.SELECTED,
                priority_winner.uid,
                priority_winner.subject,
                priority_winner.accuracy,
                priority_winner.reliability
            ),
            priority_winner
        )

    def models_with_the_biggest_reliability(self, reliability_dict):
        """Select models with the biggest reliability.

        Args:
            reliability_dict (dict): Map reliability to list of models.

        Returns:
            List:
                Pragmatic machine learning models with the biggest reliability.

        """
        biggest_reliabllity = max(reliability_dict.keys())
        winners = reliability_dict[biggest_reliabllity]
        return winners

    def models_with_the_smallest_domain(self, winners):
        """Select models with the smalles feature set.

        Args:
            winners (list): List of pragmatic machine learning models.

        Returns:
            list:
                Pragmatic machine learning models with the smallest features set.

        """
        max_domain_length = float("inf")
        winner_group = []

        for model in winners:
            if len(model.learnblock_features) < max_domain_length:
                max_domain_length = len(model.learnblock_features)

        for model in winners:
            if len(model.learnblock_features) == max_domain_length:
                winner_group.append(model)

        return winner_group

    def models_with_the_biggest_accuracy(self, winners):
        """Select the pragmatic machine learning models with the highest accuracy.

        Args:
            winners (list): List of pragmatic machine learning models.

        Returns:
            dict:
                Map of subject (abbreviation) to pragmatic machine learning model
                with the highest accuracy.

        """
        winner_accuracy = max([m.accuracy for m in winners])
        return {m.subject: m for m in winners if m.accuracy == winner_accuracy}

    def model_with_highest_priority_winner(self, winners, ml_models):
        """Select the pragmatic machine learning model with the biggest priority.

        Args:
            winners (dict): Map subject (abbreviation) to model.
            ml_models (list): Supervised machine learnign models.

        Returns:
            PragmaticMachineLearningModel:
                Model with the highest priority accoring the order of ml_models.

        """
        winner_subjects = set(winners.keys())

        if len(winners) > 1:
            for m in ml_models:
                if m.subject in winner_subjects:
                    return winners[m.subject]

        return winners.popitem()[1]


class FeatureSelector:
    """Responsible for selecting a sub set of features.

    Args:
        filter_method (MachineLearningModel): Encapsulates filter methods.
        embedded_method (MachineLearningModel): Encapsulates embed methods.
        settings (FeatureSelectionSettings):
            Controlls behavior of the feature selection.

    """
    def __init__(self, filter_method, embedded_method, settings):
        self.filter_method = filter_method
        self.embedded_method = embedded_method
        self.settings = settings

    def __str__(self):
        return "\n".join([
                "{:20}: {}".format("Filter Method", str(self.filter_method)),
                "{:20}: {}".format("Embedded Method", str(self.embedded_method)),
                str(self.settings)
            ]
        )

    def __repr__(self):
        return ", ".join([
            "{}={}".format("Filter Method", str(self.filter_method)),
            "{}={}".format("Embedded Method", str(self.embedded_method)),
            repr(self.settings)
            ]
        )

    def select(self, lb):
        """Try sub selecting the feature set of a learnblock.

        Args:
            lb (PandasBlock): Processing learnblock.

        Returns:
            tuple:
                Logging info and if selection process succeded the reduced
                learnblock else None.

        """
        while 1:
            if self.reduce_constraint(lb):
                past_n_feature = lb.n_columns(effective=True)
                self.reduce(lb)
                if self.none_reducible_valid_learnblock(past_n_feature, lb):
                    return (
                        SelectionInfo(SelectionInfo.State.REDUCIBLE, lb),
                        lb
                    )

                elif self.none_reducible_invalid_learnblock(past_n_feature, lb):
                    return (
                        SelectionInfo(SelectionInfo.State.NON_REDUCIBLE, lb),
                        None
                    )

            else:
                return SelectionInfo(SelectionInfo.State.REDUCIBLE, lb), lb

    def reduce(self, learnblock):
        """Reduce features set of the given learnblock.

        Args:
            learnblock (PandasBlock): Processing learnblock.

        """
        if self._filter_method_criteria(learnblock.n_columns(effective=True),
                                        learnblock.n_rows()):
            remove_features = self.filtering(learnblock)
            learnblock.drop_columns(remove_features, by_index=True)

        else:
            remove_features = self.embedding(learnblock)
            learnblock.drop_columns(remove_features, by_index=True)

    def reduce_constraint(self, lb):
        """Check if the feature set should be reduced.

        Args:
            lb (PandasBlock): Processing learnblock.

        Returns:
            bool:
                True if the learnblock should be further reduced.

        """
        return (
            self._to_many_features(lb.n_columns(effective=True)) or
            self.settings.max_model_reduction and not
            lb.n_columns(effective=True) == 2
        )

    def none_reducible_valid_learnblock(self, past_n_features, lb):
        """Check if the learnblock cannot be further reduced but has
        the correct number of features.

        Args:
            past_n_features (int): Number of features before feature selection.
            lb (PandasBlock): Learnblock after feature selection.

        Returns:
            bool:
                True, learnblock cannot be further reduced and has to many
                features.

        """
        return (
            past_n_features == lb.n_columns(effective=True) and not
            self._to_many_features(lb.n_columns(effective=True))
        )

    def none_reducible_invalid_learnblock(self, past_n_features, lb):
        """Check if the learnblock cannot be further reduced and has
        still to many features.

        Args:
            past_n_features (int): Number of features before feature selection.
            lb (PandasBlock): Learnblock after feature selection.

        Returns:
            bool:
                True, learnblock cannot be further reduced and has to many
                features.

        """
        return (
            past_n_features == lb.n_columns(effective=True) and
            self._to_many_features(lb.n_columns(effective=True))
        )

    def filtering(self, learnblock):
        """Reduce feature number with embedding methods.

        Args:
            learnblock (PandasBlock): Learnblock.

        Returns:
            PandasBlock:
                PandasBlock after feature reduction.

        """
        self.filter_method.train(learnblock)
        return self.filter_method.reduce(learnblock)

    def embedding(self, learnblock):
        """Reduce features number with embedding methods.

        Args:
            learnblock (PandasBlock): Learnblock.

        Returns:
            PandasBlock:
                PandasBlock after feature reduction.

        """
        self.embedded_method.train(learnblock)
        return self.embedded_method.reduce(learnblock)

    def _to_many_features(self, learn_cols):
        """Check if learnblock contains to many features.

        Args:
            learn_cols (int): Number of learnblock features.

        Returns:
            bool:
                Number of features exceeds the allowed minimum number.

        """
        return learn_cols > self.settings.max_features

    def _filter_method_criteria(self, learn_cols, samples_count):
        """Check if the filter method criteria is fullfilled.

        Args:
            learn_cols (int): Number of features of processing learnblock.
            samples_count (int): number of samples of processing learnblock.

        Returns:
            bool:
                Number of features or number of samples exceeding filter method
                criteria.

        """
        return (
            learn_cols > self.settings.max_filter_x or
            samples_count > self.settings.max_filter_y
        )



#  C O N S T R U C T I V I S T__ __
#   _____ ____   ____   /  |/  // /
#  / ___// __ \ / __ \ / /|_/ // /
# / /__ / /_/ // / / // /  / // /___
# \___/ \____//_/ /_//_/  /_//_____/
#  M A C H I N E   L E A R N I N G
#
#
# A Project by
# Thomas Schmid | UNIVERSITÄT LEIPZIG
# www.constructivist.ml
#
# Code Author: Dmitrij Denisenko
# Licence: MIT


"""Defines request for commands and queries.

"""


from collections import namedtuple


ConstructionRequest = namedtuple(
    "ConstructionRequest", "settings, algorithms, interface, mode"
)


FeatureSelectionRequest = namedtuple("FeatureSelectionRequest",
                                     ("filter_method,"
                                      "embedded_method,"
                                      "settings"))


ReconstructionRequest = namedtuple("ReconstructionRequest", ("settings,"
                                                             "algorithms,"
                                                             "interface,"
                                                             "mode,"
                                                             "krippendorf"))

DeconstructionRequest = namedtuple("DeconstructionRequest",
                                   ("general_settings,"
                                    "deconstruction_settings,"
                                    "density_estimator,"
                                    "ckmeans,"
                                    "block_processing_settings,"
                                    "build_block_from_rows,"
                                    "build_new_learnblock,"
                                    "reconstructor"))


KnowledgeSearchingRequest = namedtuple("KnowledgeSearchingRequest",
                                       ("general_settings,"
                                        "block_processing_settings,"
                                        "density_estimator,"
                                        "df_converter,"
                                        "build_block_from_rows,"
                                        "ckmeans,"
                                        "constructor,"
                                        "feature_selector,"
                                        "reconstructor,"
                                        "deconstructor,"
                                        "knowledge,"
                                        "stdout,"
                                        "n_procs"))



#  C O N S T R U C T I V I S T__ __
#   _____ ____   ____   /  |/  // /
#  / ___// __ \ / __ \ / /|_/ // /
# / /__ / /_/ // / / // /  / // /___
# \___/ \____//_/ /_//_/  /_//_____/
#  M A C H I N E   L E A R N I N G
#
#
# A Project by
# Thomas Schmid | UNIVERSITÄT LEIPZIG
# www.constructivist.ml
#
# Code Author: Dmitrij Denisenko
# Licence: MIT


"""Define the knowledge database.

"""


import pickle
from bisect import bisect
from contextlib import suppress
from collections import defaultdict

from conML.shared.errors import IndexOutOfBucket


def notify_inverted_index(func):
    def wrapper(self, *args, **kwargs):
        for obs in self.observer:
            getattr(obs, func.__name__)(*args, **kwargs)
        return func(self, *args, **kwargs)
    return wrapper


class KnowledgeDatabase:
    """Responsible for holding all models that are saved during the
    training.

    Args:
        highest_level (int): The maximal level of the database.
        build_block_from_rows (Callable): Interface to create a
            PandasBlock from pandas.Dataframe samples.

    Attributes:
        database (list): List the KnowledgeDomains.
        source (KnowledgeSource): Holds valid learnblocks.
        n_models (int): Number of models int the database.
        observer (RelativeFinder): Object that monitors changes done
            to the database.


    """

    def __init__(self, highest_level, build_block_from_rows):
        self.highest_level = highest_level
        self.database = [KnowledgeDomain(i) for i in range(highest_level+1)]
        self.source = KnowledgeSource(build_block_from_rows)
        self.observer = []
        self.n_models = 0

    def __str__(self):
        string = "{}".format(str(self.source))
        for domain in self.database:
            string = "\n".join([string, str(domain)])

        return string

    def __repr__(self):
        string = "N_models={}, Source={}".format(str(self.n_models), repr(self.source))
        for domain in self.database:
            domain_repr = repr(domain)
            string = " ".join([
                string, "[{}={}]".format("KnowledgeDomain", domain_repr)
                ]
            )
        return string

    def __contains__(self, item):
        with suppress(KeyError):
            self.get_model(item.uid)
            return True

        return False

    def extend(self, block):
        """Save the learnblock into knowledge database.

        Args:
            block (PandasBlock): Valid learnblock.

        Returns:
            int:
                Number of the inserted block.

        """
        return self.source.append(block)

    def get_block(self, indices):
        """Get learnblock by indices.

        Args:
            indices (list): List with valid indices.

        Returns:
            PandasBlock:
                Constructed from the indices.

        """
        return self.source.get(indices)

    def remove_dependent(self, model):
        """Remove the given model from the database with all other models that
        depend on it.

        Args:
            model (PragmaticMachineLearningModel): Model to be removed.

        Returns:
            List:
                List the deleded models.

        """
        if model.level == self.highest_level:
            self.remove(model)
            return [model]

        else:
            to_delete = [[model]]
            for i in range(self.highest_level):
                d = []
                for m in to_delete[i]:
                    d.extend(list(self.find_dependence_on_next_level(m)))
                to_delete.append(d)

            flatted = set([item for lst in to_delete for item in lst])
            for candidate in flatted:
                try:
                    self.remove(candidate)
                except KeyError:
                    continue

            return flatted

    @notify_inverted_index
    def insert(self, model):
        """Insert model into knowledge database.

        Args:
            model (PragmaticMachineLearningModel): Model to be inserted.

        """
        if model not in self:
            self.database[model.level].insert(model)
            self.n_models += 1

    @notify_inverted_index
    def remove(self, model):
        """Remove model from knowledge database.

        Args:
            model (PragmaticMachineLearningModel): Model to be removed.

        """
        self.database[model.level].remove(model)
        self.n_models -= 1

    def get_model(self, uid):
        """Return model by uid.

        Args:
            uid (str): Unique model identifier.

        Returns:
            PragmaticMachineLearningModel

        """
        _, level, _ = uid.split(".")
        return self.database[int(level)].get(uid)

    def save(self, path):
        """Serialize the knowledge database to the harddrive.

        Args:
            path (str): Storage destination.

        """
        with open(path, "wb") as file:
            pickle.dump(self, file)

    def find_dependence_on_next_level(self, model):
        """Yield all dependencies model on the next higher level.

        Args:
            model (PragmaticMachineLearningModel): On which the dependencies
                are searched.

        Yields:
            PragmaticMachineLearnignModel:
                Dependent on the given model.

        """
        if model.level == self.highest_level:
            yield model
        else:
            domain = self.database[model.level+1]
            for m in domain.knowledge.values():
                if model.uid in m.origin:
                    yield m

    def get_next_identifier(self, level):
        """Return valid identifier for a PragmaticMachineLearningModel.

        Args:
            level (int): Database level.

        Returns:
            int:
                Valid identifier.

        """
        return self.database[level].next_id


class KnowledgeDomain:
    """Representing a domain within a knowledge database.

    Args:
        level (int): Domain level.

    Attributes:
        knowledge (dict): Storage for PragmaticMachineLearningModel.
        biggest_id (int): Largest assigned identifier.

    """

    def __init__(self, level):
        self.level = level
        self.knowledge = {}
        self.biggest_id = 0

    def __str__(self):
        pragmatics = ", ".join([str(v.uid) for k, v in self.knowledge.items()])
        return "\n".join([
                "{:*^100}".format("Level {}".format(str(self.level))),
                "{:20}: {}".format("Models", pragmatics if pragmatics else []),
            ]
        )

    def __repr__(self):
        pragmatics = ", ".join([str(v.uid) for k, v in self.knowledge.items()])
        return ", ".join([
                "{}={}".format("Level", self.level),
                "{}={}".format("Models", pragmatics if pragmatics else []),
                "{}={}".format("Biggest ID", self.biggest_id)
            ]
        )

    @property
    def next_id(self):
        """Return next valid identifier.

        Returns:
            int:
                Valid identifier.

        """
        self.biggest_id += 1
        return self.biggest_id

    def get(self, uid):
        """Return model with the given uid.

        Args:
            uid (str): Unique model identifer.

        Returns:
            PragmaticMachineLearningModel

        """
        return self.knowledge[uid]

    def insert(self, model):
        """Insert model on this domain.

        Args:
            model (PragmaticMachinelearningModel): To be inserted.

        """
        self.knowledge[model] = model

    def remove(self, model):
        """Remove given model from this domain.

        Args:
            model (PragmaticMachineLearnignModel): To be removed.

        """
        del self.knowledge[model]


class KnowledgeSource:
    """Responsible for storing valid learnblocks. So they can be
    restored using indices.

    Args:
        pandas_factory (Callable): Builds PandasBlocks from indices.

    Attributes:
        source (dict): Maps block number to learnblock.
        n_blocks (int): Counts the inserted blocks.
        __last_index (int): Last index in the last block.
        __search_ranges (list): Last indecis of every stored block.

    """
    def __init__(self, pandas_factory):
        self.source = {}
        self.n_block = 0
        self.pandas_factory = pandas_factory
        self.__last_index = 0
        self.__search_ranges = []

    def __str__(self):
        return "\n".join([
                "{:20}: {}".format("# Blocks", str(self.n_block)),
            ]
        )

    def __repr__(self):
        range_repr = {}
        for i, range_end in enumerate(self.__search_ranges):
            if i == 0:
                range_repr[i] = {"range": (0, range_end)}

            else:
                range_repr[i] = {
                    "range": (self.__search_ranges[i-1], range_end)
                }

        for k, v in self.source.items():
            range_repr[k]["shape"] = v.shape()

        return ", ".join([
                "{}={}".format("Blocks", str(self.n_block)),
                "{}={}".format("Sources", range_repr)
            ]
        )

    def append(self, block):
        """Store block into dictionary.

        After storing the block, update the indices to be consistent with
        every other stored block.

        Args:
            block (PandasBlock):  Valid learnblock.

        Returns:
            int:
                Number of the stored block.

        """
        self.source[self.n_block] = block
        self.__reindex(block)
        self.__update_search_range()
        self.n_block += 1
        return self.n_block - 1

    def get(self, indices):
        """Get block via indices.

        Args:
            indices (list): List of indices.

        Returns:
            PandasBlock:
                Reconstructed from given indices.

        """
        rows = list(self.indices_to_rows(indices))
        return self.pandas_factory(rows)

    def __reindex(self, block):
        """Adjust the indices of the given block, so that the the new indices
        will start, where idices of the last stored block ended.

        Args:
            block (PandasBlock): Valid learnblock.

        """
        new_end_indx = self.__last_index+block.n_rows()
        block.reindex(range(self.__last_index, new_end_indx))
        self.__last_index = new_end_indx

    def __update_search_range(self):
        """Append the last index of the last stored block.

        """
        self.__search_ranges.append(self.__last_index)

    def indices_to_rows(self, indices):
        """Find the corresponding row given given a list of indices.

        Args:
            indices (list): List of integers representing row indices of a
                learnblock.

        Yields:
            pandas.DataFrame:
                Representing a single row.

        """
        for index in indices:
            bucket_index = self.bucket_search(index)
            yield self.source[bucket_index].row_as_df_via_index(index)

    def bucket_search(self, index):
        """Search the number of block in which the row with the given index
        is stored.

        Args:
            index (int): Rows index.

        Returns:
            int:
                Block number.

        """
        idx = bisect(self.__search_ranges, index, lo=0, hi=self.n_block-1)
        if idx >= len(self.__search_ranges):
            raise IndexOutOfBucket(
                "Row index out of bucket range."
                "Returns the bucket index, in which rows stored."
                "Make sure the row indices are stored in the corresponding "
                "bucket."
            )

        return idx


class RelativeFinder:
    """Inverted index.

    Maps meta data (timestamps, subject and aim) to models.

    Args:
        levels (int): Levels used in the knowledge database.

    """
    def __init__(self, levels):
        self.indices = {}
        for level in range(levels+1):
            self.indices[level] = (defaultdict(set),
                                   defaultdict(set),
                                   defaultdict(set))

    def __repr__(self, *args):
        level = args[0]
        return (
            f"T: {self.indices[level][0]} \n"
            f"Z: {self.indices[level][1]} \n"
            f"Sigma: {self.indices[level][2]} \n"
        )

    def __contains__(self, item):
        t, z, s = self.get_index_level(item)

        in_t = False
        if (item.min_timestamp, item.max_timestamp) in t.keys():
            in_t = item in t[(item.min_timestamp, item.max_timestamp)]

        in_z = False
        if item.aim in z.keys():
            in_z = item in z[item.aim]

        in_s = []
        for subject in item.subjects:
            if subject in s.keys():
                in_s.append(item in s[subject])

        return in_t and in_z and all(in_s)

    def find(self, pair, model, deconst_full_tolerance):
        """Find the all relatives for the given model.

        Args:
            pair (tuple): Relationship.
            model (PragmaticMachineLearningModel): Instantiaed valid model.
            deconst_full_tolerance (float): How big the distance is allowed
                to be in a complete relationship.

        Returns:
            list:
                Sorted list of relatives according their uid..

        """
        index_t, index_z, index_sigma = self.get_index_level(model)
        relatives = {}

        if pair == ("T", "Z"):
            set_t = index_t[(model.min_timestamp, model.max_timestamp)]
            set_z = index_z[model.aim]
            relatives = set_t.intersection(set_z)

        elif pair == ("T", "Sigma"):
            set_t = index_t[(model.min_timestamp, model.max_timestamp)]
            set_s = set()
            for subject in model.subjects:
                set_s = set_s.union(index_sigma[subject])
            relatives = set_t.intersection(set_s)

        elif pair == ("Sigma", "Z"):
            set_s = set()
            for subject in model.subjects:
                set_s = set_s.union(index_sigma[subject])

            set_z = index_z[model.aim]
            relatives = set_s.intersection(set_z)

        elif pair == ("complete", ):
            set_t = set()
            for min_time, max_time in index_t.keys():
                if self.complete_relationship_constraint(
                        model, min_time, max_time, deconst_full_tolerance
                ):
                    set_t.update(index_t[(min_time, max_time)])
            set_s = set()
            for subject in model.subjects:
                set_s = set_s.union(index_sigma[subject])
            set_z = index_z[model.aim]
            relatives = set_t.intersection(set_s).intersection(set_z)

        if model in relatives:
            relatives.remove(model)

        return sorted(relatives, key=lambda m: m.identifier)

    def complete_relationship_constraint(
            self, model, min_t, max_t, deconst_full_tolerance
    ):
        """Check if the model is fullfilling the complete relationship.

        That is, if the time values of the main model are within the given
        min_t and max_t or enclosing these values.

        Args:
            model (PragmaticMachineLearningModel): Instantiaded valid model.
            min_t (int): Min time of the possible relative.
            max_t (int): Max time of the possible relative.
            deconst_full_tolerance (float): How big should be the enclosing
                and exclusion distance be allowed.

        Returns:
            bool:
                True if the model is fullfilling the complete relationship
                regarding his time values.

        """
        return (
            self.__time_exclusion(model, min_t, max_t, deconst_full_tolerance)
            or
            self.__time_inclusion(model, min_t, max_t, deconst_full_tolerance)
        )

    def __time_inclusion(self, model, min_t, max_t, tolerance):
        """Check if the model timestamps include the given time values
            on condition that the given distance is not exceeded.

        Args:
            model (PragmaticMachineLearningModel): Instantiaded valid model.
            min_t (int): Min time of the possible relative.
            max_t (int): Max time of the possible relative.
            tolerance (float): Distance tolerance.

        Returns:
            bool:
                True if the model time is including the given times.

        """
        return (
            model.min_timestamp >= min_t and model.max_timestamp <= max_t
            and
            model.min_timestamp - min_t <= model.min_timestamp * tolerance
            and
            model.max_timestamp - max_t <= model.max_timestamp * tolerance
        )

    def __time_exclusion(self, model, min_t, max_t, tolerance):
        """Check if the model timestamps exclude both given timestamps.

        Args:
            model (PragmaticMachineLearningModel): Instantiaded valid model.
            min_t (int): Min time of the possible relative.
            max_t (int): Max time of the possible relative.
            tolerance (float): Distance tolerance.

        Returns:
            bool:
                True if the model time is excluding both given timestamps.

        """
        return (
            model.min_timestamp <= min_t and model.max_timestamp >= max_t
            and
            model.min_timestamp - min_t <= model.min_timestamp * tolerance
            and
            model.max_timestamp - max_t <= model.max_timestamp * tolerance
        )

    def remove(self, model):
        """Remove given model from the inverted index.

        Args:
            model (PragmaticMachineLearningModel): Instantiaded valid model.

        """
        t_list, s_lists, z_list = self.get_index_lists(model,
                                                      time=True,
                                                      subject=True,
                                                      aim=True)
        for s_list in s_lists:
            if model in s_list:
                s_list.remove(model)

            if not s_list:
                del s_list

        if model in z_list:
            z_list.remove(model)

        if model in t_list:
            t_list.remove(model)

        if not t_list:
            del self.indices[model.level][0][(model.min_timestamp,
                                              model.max_timestamp)]

        if not z_list:
            del self.indices[model.level][1][model.aim]

    def insert(self, model):
        """Insert given model into the inverted index.

        Args:
            model (PragmaticMachineLearningModel): Instantiaded valid model.

        """
        t, z, sigma = self.get_index_level(model)
        t[(model.min_timestamp, model.max_timestamp)].add(model)
        z[model.aim].add(model)

        for subject in model.subjects:
            sigma[subject].add(model)

    def get_index_lists(self, model, time, subject, aim):
        """Get all models whose meta data equals the model timestamps,
        subject and aim.

        Args:
            model (PragmaticMachineLearningModel): Instantiaded valid model.
            time (bool): Indicates if time relatives should be found.
            subject (bool): Indicates if the subject relatives should be found.
            aim (bool): Indicates if the aim relatives should be found.

        Returns:
            tuple:
                Lists with relatives.

        """
        t, z, sigma = self.get_index_level(model)
        t_list = s_list = z_list = None
        if time:
            t_list = t[(model.min_timestamp, model.max_timestamp)]
        if subject:
            s_lists = []
            for subject in model.subjects:
                s_lists.append(sigma[subject])
            # s_list = sigma[model.subject]
        if aim:
            z_list = z[model.aim]

        return t_list, s_lists, z_list

    def get_index_level(self, model):
        """Get time, subject and aim index on model level.

        Args:
            model (PragmaticMachineLearningModel): Instantiaded valid model.

        Returns:
            tuple:
                List with indices.

        """
        index_t = self.indices[model.meta.knowledge_level][0]
        index_z = self.indices[model.meta.knowledge_level][1]
        index_sigma = self.indices[model.meta.knowledge_level][2]
        return index_t, index_z, index_sigma



#  C O N S T R U C T I V I S T__ __
#   _____ ____   ____   /  |/  // /
#  / ___// __ \ / __ \ / /|_/ // /
# / /__ / /_/ // / / // /  / // /___
# \___/ \____//_/ /_//_/  /_//_____/
#  M A C H I N E   L E A R N I N G
#
#
# A Project by
# Thomas Schmid | UNIVERSITÄT LEIPZIG
# www.constructivist.ml
#
# Code Author: Dmitrij Denisenko
# Licence: MIT


"""Define single access point to all control parameter.

Every parameter is set to its default value after reading the configuration
file. The default configuration file is located at $HOME/.conML/settings.ini on
Unix like OS and it is read during conML module import by the user.
It is possible to provide another settings.ini path destination.

"""


from itertools import starmap
from dataclasses import dataclass, asdict
from configparser import ConfigParser

from conML.shared.parameter import *
from conML.shared.errors import ConfigurationFileError


class MetaSettings(type):
    """Container for all parameter.

    """
    HIGHEST_LEVEL: int = HighestLevel()
    KERNEL_BANDWIDTH: float = KernelBandwidth()

    LEARN_BLOCK_MINIMUM: int = LearnblockMinimum()
    SIGMA_ZETA_CUTOFF: float = SigmaZetaCutoff()

    MAX_CATEGORIES: int = MaxCategories()
    MIN_CATEGORY_SIZE: int = MinCategorySize()
    MAX_MODEL_TARGETS: int = MaxModelTargets()
    MAX_TARGET_ERROR: float = MaxTargetError()

    MIN_BUILD_MODELS: int = MinBuildModels()
    MAX_FEATURES: int = MaxFeatures()
    MAX_FILTER_X: int = MaxFilterX()
    MAX_FILTER_Y: int = MaxFilterY()
    MAX_MODELS_REDUCTION: bool = MaxModelsReduction()
    MIN_TEST_ACCURACY: float = MinTestAccuracy()
    MAX_TEST_ERROR_AVG: float = MaxTestErrorAvg()
    MAX_TEST_ERROR_MAX: float = MaxTestErrorMax()
    RELIABILITY_SAMPLE: float = ReliabilitySample()
    MIN_RELIABILITY: float = MinReliability()
    REDUCE_MODEL_REDUNDANCY: bool = ReduceModelRedundancy()

    DECONST_STRATEGY: str = DeconstStrategy()
    DECONST_MODE: str = DeconstMode()
    DECONST_MAX_DISTANCE_T: int = DeconstMaxDistanceT()
    DECONST_FULL_TOLERANCE: float = DeconstFullTolerance()
    FORCE_TIME_EXPANSION: bool = ForceTimeExpansion()
    ALLOW_WEAK_RELIABILITY: bool = AllowWeakReliability()


class Settings(metaclass=MetaSettings):
    pass


@dataclass
class GeneralSettings:
    """Influence the global framework.

    """
    highest_level: int = HighestLevel()
    kernel_bandwidth: float = KernelBandwidth()

    def __str__(self):
        return "\n".join([
            "{:<20}: {}".format("Highest level", self.highest_level)
        ])


@dataclass
class BlockProcessingSettings:
    """Influence the block drawing.

    """
    learn_block_minimum: int = LearnblockMinimum()
    sigma_zeta_cutoff: float = SigmaZetaCutoff()
    min_category_size: int = MinCategorySize()

    def __str__(self):
        return "\n".join([
            "{:<20}: {}".format("Learn_block_minimum", self.learn_block_minimum),
            "{:<20}: {}".format("Sigma zeta cutoff", self.sigma_zeta_cutoff),
        ])


@dataclass
class ConstructionSettings:
    """Influence the construction step.

    """
    max_categories: int = MaxCategories()
    min_category_size: int = MinCategorySize()
    max_model_targets: int = MaxModelTargets()
    max_target_error: float = MaxTargetError()

    def __str__(self):
        return "\n".join([
            "{:<20}: {}".format("Max categories", self.max_categories),
            "{:<20}: {}".format("Min category size", self.min_category_size),
            "{:<20}: {}".format("Max model targets", self.max_model_targets),
            "{:<20}: {}".format("Max target error", self.max_target_error),
        ])

    def __repr__(self):
        return ", ".join([
            "{}={}".format("max_categories", self.max_categories),
            "{}={}".format("min_category size", self.min_category_size),
            "{}={}".format("max_model_targets", self.max_model_targets),
            "{}={}".format("max_target_error", self.max_target_error),
        ])


@dataclass
class FeatureSelectionSettings:
    """Influence the feature selection step.

    """
    max_features: int = MaxFeatures()
    max_filter_x: int = MaxFilterX()
    max_filter_y: int = MaxFilterY()
    max_model_reduction: bool = MaxModelsReduction()

    def __str__(self):
        return "\n".join([
            "{:<20}: {}".format("Max features", self.max_features),
            "{:<20}: {}".format("Filter X", self.max_filter_x),
            "{:<20}: {}".format("Filter Y", self.max_filter_y),
            "{:<20}: {}".format("Model Reduction", self.max_model_reduction),
            ""
        ])


@dataclass
class ReconstructionSettings:
    """Influence the reconstruction step.

    """
    min_test_accuracy: float = MinTestAccuracy()
    max_test_error_avg: float = MaxTestErrorAvg()
    max_test_error_max: float = MaxTestErrorMax()
    reliability_sample: float = ReliabilitySample()
    min_reliability: float = MinReliability()
    reduce_model_redundancy: bool = ReduceModelRedundancy()
    min_build_models: int = MinBuildModels()

    def __str__(self):
        return "\n".join([
            "{:<20}: {}".format("Min test accuracy", self.min_test_accuracy),
            "{:<20}: {}".format("Max test error avg", self.max_test_error_avg),
            "{:<20}: {}".format("Max test error max", self.max_test_error_max),
            "{:<20}: {}".format("Reliability sample", self.reliability_sample),
            "{:<20}: {}".format("Reliability sample", self.min_reliability),
            "{:<20}: {}".format("Reduce model rudendancy", self.reduce_model_redundancy),
            "{:<20}: {}".format("Min build models", self.min_build_models)
        ])


@dataclass
class DeconstructionSettings:
    """Influence the deconstruction step.

    """
    deconst_strategy: str = DeconstStrategy()
    deconst_mode: str = DeconstMode()
    deconst_max_distance_t: int = DeconstMaxDistanceT()
    deconst_full_tolerance: float = DeconstFullTolerance()
    force_time_expansion: bool = ForceTimeExpansion()
    allow_weak_reliability: bool = AllowWeakReliability()
    learn_block_minimum: int = LearnblockMinimum
    min_reliability: float = MinReliability()

    def __str__(self):
        return "\n".join([
            "{:<20}: {}".format("Deconst strategy", self.deconst_strategy),
            "{:<20}: {}".format("Deconst mode", self.deconst_mode),
            "{:<20}: {}".format("Deconst max distance t", self.deconst_max_distance_t),
            "{:<20}: {}".format("Deconst full tolerance", self.deconst_full_tolerance),
            "{:<20}: {}".format("Force time expansion", self.force_time_expansion),
            "{:<20}: {}".format("Allow weak reliability", self.allow_weak_reliability),
            "{:<20}: {}".format("Learnblock minimum", self.learn_block_minimum),
            "{:<20}: {}".format("Min reliability", self.min_reliability)
        ])

    def as_dict(self):
        return asdict(self)


def specific_settings_factory(settings_type):
    """Factory method for specific settings objects.

    Returns:
         One of the settings objects.

    """
    factory = {
        "general": starmap(
            GeneralSettings, [(Settings.HIGHEST_LEVEL,
                               Settings.KERNEL_BANDWIDTH)]),

        "block_processing": starmap(
            BlockProcessingSettings, [(Settings.LEARN_BLOCK_MINIMUM,
                                       Settings.SIGMA_ZETA_CUTOFF,
                                       Settings.MIN_CATEGORY_SIZE)]),

        "construction": starmap(
            ConstructionSettings, [(Settings.MAX_CATEGORIES,
                                    Settings.MIN_CATEGORY_SIZE,
                                    Settings.MAX_MODEL_TARGETS,
                                    Settings.MAX_TARGET_ERROR)]),

        "feature_selection": starmap(
            FeatureSelectionSettings, [(Settings.MAX_FEATURES,
                                        Settings.MAX_FILTER_X,
                                        Settings.MAX_FILTER_Y,
                                        Settings.MAX_MODELS_REDUCTION)]),

        "reconstruction": starmap(
            ReconstructionSettings, [(Settings.MIN_TEST_ACCURACY,
                                      Settings.MAX_TEST_ERROR_AVG,
                                      Settings.MAX_TEST_ERROR_MAX,
                                      Settings.RELIABILITY_SAMPLE,
                                      Settings.MIN_RELIABILITY,
                                      Settings.REDUCE_MODEL_REDUNDANCY,
                                      Settings.MIN_BUILD_MODELS)]),

        "deconstruction": starmap(
            DeconstructionSettings, [(Settings.DECONST_STRATEGY,
                                      Settings.DECONST_MODE,
                                      Settings.DECONST_MAX_DISTANCE_T,
                                      Settings.DECONST_FULL_TOLERANCE,
                                      Settings.FORCE_TIME_EXPANSION,
                                      Settings.ALLOW_WEAK_RELIABILITY,
                                      Settings.LEARN_BLOCK_MINIMUM,
                                      Settings.MIN_RELIABILITY)])
    }

    return next(factory[settings_type])


def read_settings(path):
    """Read settings from configuration file and assign their values
    to the global Settings object.

    Args:
        path (str): Path to the configuration file.

    raises:
        ConfigurationFileError:
            Settings from the configuration file cannot be read.

    """
    try:
        config = ConfigParser()
        config.read(path)
        configure_main_settings_class(config)

    except AttributeError as e:
        raise ConfigurationFileError(
            "Error during reading the configuration file."
            "Tried read the configuration file."
            "Make sure that all parameters are in the configuration file."
        )


def configure_main_settings_class(config):
    """Assign the values from the configuration file to the global
    Settings object.

    Args:
        config (dict): Read configuration file.

    """
    default = config["GENERAL"]
    Settings.HIGHEST_TIER = default["highest_level"]

    block_processing = config["BLOCK_PROCESSING"]
    Settings.LEARN_BLOCK_MINIMUM = block_processing["learn_block_minimum"]
    Settings.SIGMA_ZETA_CUTOFF = block_processing["sigma_zeta_cutoff"]

    construction = config["CONSTRUCTION"]
    Settings.MAX_TARGET_ERROR = construction["max_target_error"]
    Settings.MAX_MODEL_TARGETS = construction["max_model_targets"]
    Settings.MAX_CATEGORIES = construction["max_categories"]
    Settings.MIN_CATEGORY_SIZE = construction["min_category_size"]

    feature_selection = config["FEATURE_SELECTION"]
    Settings.MAX_FEATURES = feature_selection["max_features"]
    Settings.MAX_FILTER_X = feature_selection["max_filter_x"]
    Settings.MAX_FILTER_Y = feature_selection["max_filter_y"]
    Settings.MAX_MODELS_REDUCTION = feature_selection["max_model_reduction"]

    reconstruction = config["RECONSTRUCTION"]
    Settings.MIN_TEST_ACCURACY = reconstruction["min_test_accuracy"]
    Settings.MAX_TEST_ERROR_AVG = reconstruction["max_test_error_avg"]
    Settings.MAX_TEST_ERROR_MAX = reconstruction["max_test_error_max"]
    Settings.RELIABILITY_SAMPLE = reconstruction["reliability_sample"]
    Settings.MIN_RELIABILITY = reconstruction["min_reliability"]
    Settings.REDUCE_MODEL_REDUNDANCY = reconstruction["reduce_model_redundancy"]
    Settings.MIN_BUILD_MODELS = reconstruction["min_build_models"]

    deconstruction = config["DECONSTRUCTION"]
    Settings.DECONST_STRATEGY = deconstruction["deconst_strategy"]
    Settings.DECONST_MODE = deconstruction["deconst_mode"]
    Settings.DECONST_MAX_DISTANCE_T = deconstruction["deconst_max_distance_t"]
    Settings.DECONST_FULL_TOLERANCE = deconstruction["deconst_full_tolerance"]
    Settings.FORCE_TIME_EXPANSION = deconstruction["force_time_expansion"]
    Settings.ALLOW_WEAK_RELIABILITY = deconstruction["allow_weak_reliability"]



#  C O N S T R U C T I V I S T__ __
#   _____ ____   ____   /  |/  // /
#  / ___// __ \ / __ \ / /|_/ // /
# / /__ / /_/ // / / // /  / // /___
# \___/ \____//_/ /_//_/  /_//_____/
#  M A C H I N E   L E A R N I N G
#
#
# A Project by
# Thomas Schmid | UNIVERSITÄT LEIPZIG
# www.constructivist.ml
#
# Code Author: Dmitrij Denisenko
# Licence: MIT


"""Define the deconstruction process.

"""

from collections import deque
from multiprocessing import Event, Manager
from queue import Empty

from conML.shared.logger import DeconstructionInfo
from conML.domain.deconstruction.z_sigma import ZSigmaDeconstructor
from conML.domain.deconstruction.complete import CompleteDeconstructor
from conML.domain.deconstruction.t_sigma import (
    TSigmaDeconstructor,
    Start,
    Apply,
    Upgrade,
    Clean,
    Exit
)


__all__ = (
    "Deconstructor",
)


class AbstractDeconstructor:
    """Responsible for defining basis functions that are used in the
    most deconstructions.

    Args:
        general_settings (GeneralSettings): Controlls the behavior of
            the deconstructions.
        block_processing_settings (BlockProcessingSettings): Controlls the
            behavior of deconstruction.
        deconstruction_settings (DeconstructionSettings): Controlls the
            behavior of deconstruction.
        reconstructor (Reconstructor): Reconstructor.
        learnblock_constructor (Callable): Build learnblocks.
        cluster_finder (ClusterFinder): Find cluster within time stamps.


    """
    def __init__(self, general_settings, block_processing_settings,
                 deconstruction_settings, reconstructor,
                 learnblock_constructor, cluster_finder):
        self.general_settings = general_settings
        self.block_processing_settings = block_processing_settings
        self.settings = deconstruction_settings
        self.reconstructor = reconstructor
        self.learnblock_constructor = learnblock_constructor
        self.cluster_finder = cluster_finder
        self.tsigma_beamer = deque()

    def __str__(self):
        return "\n".join([
                str(self.settings),
                str(self.reconstructor)
            ]
        )

    def __repr__(self):
        return ", ".join([
                "{}".format(repr(self.settings)),
                "{}".format(repr(self.reconstructor))
            ]
        )

    def to_dict(self):
        return {
            "general_settings": self.general_settings,
            "block_processing_settings": self.block_processing_settings,
            "deconstruction_settings": self.settings,
            "reconstructor": self.reconstructor,
            "learnblock_constructor": self.learnblock_constructor,
            "cluster_finder": self.cluster_finder
        }

    def deconstruct(self, *args, **kwargs):
        raise NotImplementedError

    def find_relatives(self, database, pragmatic):
        """Find all relatives of the given model in knowledge database.

        Args:
            database (KnowledgeDatabase): Database for
                PragmaticMachineLearningModels.
            pragmatic (PragmaticMachineLearningModel): From which the relatives
                to find.

        Yields:
            PragmaticMachineLearningModel:
                Relative.

        """
        relative_finder = database.observer[0]

        complete_relatives = relative_finder.find(
            tuple(["complete", ]), pragmatic,
            self.settings.deconst_full_tolerance)

        t_sigma_relatives = relative_finder.find(
            tuple(["T", "Sigma"]), pragmatic,
            self.settings.deconst_full_tolerance)

        sigma_z_relatives = relative_finder.find(
            tuple(["Sigma", "Z"]), pragmatic,
            self.settings.deconst_full_tolerance)

        D["complete"] = len(complete_relatives)
        yield complete_relatives

        t_sigma_relatives = [m for m in t_sigma_relatives
                             if m in relative_finder]

        D["tsigma"] = len(t_sigma_relatives)
        yield t_sigma_relatives

        sigma_z_relatives = [m for m in sigma_z_relatives
                             if m in relative_finder]
        D["sigmaz"] = len(sigma_z_relatives)
        yield sigma_z_relatives

    def clean_ts_queue(self, deleted):
        """Remove invalid models from  T-Sigma queue.

        Args:
            deleted (list): List of deleted PragmaticMachineLearningModels
                in complete deconstructino process.

        """
        queue_copy = self.tsigma_beamer.copy()
        for level, lb in queue_copy:
            for model in deleted:
                if model in lb.origin:
                    self.tsigma_beamer.remove((level, lb))


class SerialDeconstructor(AbstractDeconstructor):
    """Encapsulates the logic for the seriel deconstruction.

    Note:
        Please see help(AbstractDeconstructor) for more info!

    """
    def __str__(self):
        return super().__str__()

    def __repr__(self):
        return super().__repr__()

    def deconstruct(self, level, pragmatic, database):
        """Start the deconstruction process.

        Args:
            level (int): Level on which the deconstruction is performed.
            pragmatic (PragmaticMachineLearningModel): Recently reconstructed
                model.
            database (KnowledgeDatabase): Database to store models.

        Yield:
            DeconstructionInfo:
                Logging information about the deconstruction process.

        """
        relatives = iter(self.find_relatives(database, pragmatic))

        yield self.run_complete_deconstruction(
            level, pragmatic, database, next(relatives)
        )

        yield self.run_t_sigma_deconstruction(
            level, pragmatic, database, next(relatives)
        )

        yield self.run_sigma_z_deconstruction(
            level, pragmatic, database, next(relatives)
        )

    def run_complete_deconstruction(self, level, pragmatic, database, relatives):
        """Perform complete deconstruction.

        Args:
            level (int): Level on which the deconstruction is performed.
            pragmatic (PragmaticMachineLearningModel): Model to deconstruct.
            database (KnowledgeDatabase): Database to store models.
            relatives (list): List of 'complete' relatives.

        Yield:
            DeconstructionInfo:
                Logging information about the deconstruction process.

        """
        if relatives:
            result_deque = deque()
            CompleteDeconstructor(
                level, database, self.settings, self.reconstructor, pragmatic,
                relatives, result_deque, cluster_finder=self.cluster_finder,
                general_settings=self.general_settings,
                block_processing_settings=self.block_processing_settings
            ).run()

            for log in result_deque:
                self.clean_ts_queue(log.deleted)
                if log.state == DeconstructionInfo.State.DECONSTRUCTED:
                    D["c_deconstructed"] += 1

                yield log

    def run_t_sigma_deconstruction(self, level, pragmatic, database, relatives):
        """Perform T Sigma deconstruction.

        Args:
            level (int): Level on which the deconstruction is performed.
            pragmatic (PragmaticMachineLearningModel): Model to deconstruct.
            database (KnowledgeDatabase): Database to store models.
            relatives (list): List of 'T Sigma' relatives.

        Yield:
            DeconstructionInfo:
                Logging information about the deconstruction process.

        """
        if relatives:
            result_deque = deque()
            TSigmaDeconstructor(
                self.learnblock_constructor,
                self.reconstructor.krippendorff_alpha, database, self.settings,
                self.general_settings
            ).run_serial(level, relatives, pragmatic, result_deque)

            for log, level_, learnblock in result_deque:
                self.tsigma_beamer.append((level_, learnblock))
                yield log

    def run_sigma_z_deconstruction(self, level, pragmatic, database, relatives):
        """Perform Sigma Z deconstruction.

        Args:
            level (int): Level on which the deconstruction is performed.
            pragmatic (PragmaticMachineLearningModel): Model to deconstruct.
            database (KnowledgeDatabase): Database to store model.
            relatives (list): List of Sigma Z relatives.

        Yield:
            DeconstructionInfo:
                Logging information about the deconstruction process.

        """
        if relatives:
            result_deque = deque()
            ZSigmaDeconstructor(
                level, database, self.settings, self.reconstructor, pragmatic,
                relatives, result_deque
            ).run()

            for log in result_deque:
                self.clean_ts_queue(log.deleted)
                yield log


class ParallelDeconstructor(AbstractDeconstructor):
    """Encapsulates the logic for the parallel deconstruction.

    Note:
        Please help(AbstractDeconstructor) for more info!

    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Used if parallelization is on
        self.tsigma_process = None
        self.tsigma_inputs = None
        self.tsigma_outputs = None
        self.upgrade_event = None
        self.apply_event = None
        self.clear_event = None

    def __str__(self):
        return super().__str__()

    def __repr__(self):
        return super().__repr__()

    def deconstruct(self, level, pragmatic, database):
        """Start the deconstruction process.

        Args:
            level (int): Level on which the deconstruction is performed.
            pragmatic (PragmaticMachineLearningModel): Recently reconstructed
                model.
            database (KnowledgeDatabase): Database to store models.

        Yields:
            DeconstructionInfo:
                Logging information about the deconstruction process.

        """
        relatives = iter(self.find_relatives(database, pragmatic))
        complete_relatives = next(relatives)
        t_sigma_relavites = next(relatives)

        if t_sigma_relavites:
            self.start_background_tsigma(database)
            self.tsigma_inputs.put(Start(level=level,
                                         relatives=t_sigma_relavites,
                                         pragmatic=pragmatic))
        if complete_relatives:
            yield self.run_complete_deconstruction(
                level, pragmatic, database, complete_relatives
            )

        if t_sigma_relavites:
            yield self.run_t_sigma_deconstruction(database, pragmatic)

        yield self.run_sigma_z_deconstruction(
            level, pragmatic, database, next(relatives)
        )

    def start_background_tsigma(self, database):
        """Start the T Sigma deconstruction in a separate process.

        Args:
            database (KnowledgeDatabase): Database to store models.

        """
        if not self.tsigma_process:
            m = Manager()
            self.tsigma_inputs = m.Queue()
            self.tsigma_outputs = m.Queue()
            self.upgrade_event = Event()
            self.apply_event = Event()
            self.clear_event = Event()
            self.tsigma_process = TSigmaDeconstructor(
                self.learnblock_constructor,
                self.reconstructor.krippendorff_alpha, database, self.settings,
                self.general_settings, self.tsigma_inputs, self.tsigma_outputs,
                self.upgrade_event, self.apply_event, self.clear_event)

            self.tsigma_process.daemon = True
            self.tsigma_process.start()

    def run_complete_deconstruction(self, level, pragmatic, database, relatives):
        """Perform complete deconstruction.

        Args:
            level (int): Level on which the deconstruction is performed.
            pragmatic (PragmaticMachineLearningModel): Model to deconstruct.
            database (KnowledgeDatabase):  Database to store models.
            relatives (list): List of 'complete' relatives.

        Yields:
            DeconstructionInfo:
                Logging information about the deconstruction process.

        """
        result_deque = deque()
        CompleteDeconstructor(
            level, database, self.settings, self.reconstructor, pragmatic,
            relatives, result_deque, cluster_finder=self.cluster_finder,
            general_settings=self.general_settings,
            block_processing_settings=self.block_processing_settings
        ).run()

        for log_info in result_deque:
            if log_info.state == DeconstructionInfo.State.DECONSTRUCTED:
                D["c_deconstructed"] += 1

            if self.tsigma_process:
                self.tsigma_inputs.put(Upgrade(deletes=log_info.deleted))
            yield log_info

    def run_t_sigma_deconstruction(self, knowledge, pragmatic):
        """Perform T Sigma deconstruction.

        Args:
            knowledge (KnowledgeDatabase): Database to store models.
            pragmatic (PragmaticMachineLearningModels): Model to deconstruct.

        Yields:
            DeconstructionInfo:
                Logging information about the deconstruction process.

        """
        self.tsigma_inputs.put(Apply())
        self.apply_event.wait()

        while not self.tsigma_outputs.empty():
            try:
                log_info, level, lb = self.tsigma_outputs.get(False)
                self.tsigma_beamer.append((level, lb))
                knowledge.insert(pragmatic)
                log_info.inserted.append(pragmatic)
                yield log_info

            except Empty:
                continue

    def run_sigma_z_deconstruction(self, level, pragmatic, database, relatives):
        """Perform Sigma Z deconstruction.

        Args:
            level (int): Level on which the deconstruction is performed.
            pragmatic (PragmaticMachineLearningModel): Model to deconstruct.
            database (KnowledgeDatabase): Databases to store model.
            relatives (list): List of Sigma Z relatives.

        """
        if relatives:
            result_deque = deque()
            ZSigmaDeconstructor(
                level, database, self.settings, self.reconstructor, pragmatic,
                relatives, result_deque
            ).run()

            for log_info in result_deque:
                self.clean_ts_queue(log_info.deleted)
                yield log_info


class Deconstructor:
    """Responsible for holding either SerialDeconstructor or
    ParallelDeconstructor.

    Args:
        mode:
        general_settings (GeneralSettings): Controlls the behavior of the
            deconstruction.
        block_processing_settings (BlockProcessingSettings): Controlls the
            behavior of deconstruction.
        deconstruction_settings (DeconstructionSettings): Controlls the
            behavior of deconstruction.
        reconstructor (Reconstructor): Reconstructor.
        learnblock_constructor (Callable): Build learnblocks.
        cluster_finder (ClusterFinder): Find cluster within time stamps.

    """
    def __init__(self, general_settings, block_processing_settings,
                 deconstruction_settings, reconstructor, learnblock_constructor,
                 cluster_finder):

        self.state = SerialDeconstructor(
            general_settings, block_processing_settings,
            deconstruction_settings, reconstructor,
            learnblock_constructor, cluster_finder)

    def __str__(self):
        return str(self.state)

    def __repr__(self):
        return repr(self.state)

    def transit(self):
        """Change the state to SerialDeconstructor or ParallelDeconstructor.

        """
        if type(self.state) == SerialDeconstructor:
            self.state = ParallelDeconstructor(**self.state.to_dict())

        else:
            self.state = SerialDeconstructor(**self.state.to_dict())

    def deconstruct(self, level, pragmatic, database):
        """Start the deconstruction process.

        Args:
            level (int): Level on which the deconstruction is performed.
            pragmatic (PragmaticMachineLearningModel): Recently reconstructed
                model.
            database (KnowledgeDatabase): Database to store models.

        Yields:
            DeconstructionInfo:
                Logging information about the deconstruction.

        """
        try:
            changes_done_status = []
            for decon_type in self.state.deconstruct(level, pragmatic, database):
                for log in decon_type:
                    changes_done_status.append(log.state)
                    yield log

                    # First deconstruction successfully ended and deconst mode
                    # says, end the deconstruction.
                    if (log.state == DeconstructionInfo.State.DECONSTRUCTED and
                            self.deconst_mode == "minimal"):
                        return

                    # The new reconstructed pragmatic model was deleted, so end the
                    # whole deconstruction process.
                    elif pragmatic in log.deleted:
                        return

            if not changes_done_status or all(
                    map(
                        lambda x: x == DeconstructionInfo.State.FAILED,
                        changes_done_status
                    )
            ):
                database.insert(pragmatic)
                yield DeconstructionInfo(pragmatic=pragmatic,
                                         inserted=[pragmatic])
        finally:
            if (isinstance(self.state, ParallelDeconstructor) and
                    self.state.tsigma_process):
                self.state.tsigma_inputs.put(Clean())
                self.state.clear_event.wait()
                self.state.tsigma_inputs.put(Exit())
                self.state.tsigma_process = None

    @property
    def tsigma_beamer(self):
        return self.state.tsigma_beamer

    @property
    def deconst_mode(self):
        return self.state.settings.deconst_mode

    @deconst_mode.setter
    def deconst_mode(self, value):
        self.state.settings.deconst_mode = value

    @property
    def deconst_strategy(self):
        return self.state.settings.deconst_strategy

    @deconst_strategy.setter
    def deconst_strategy(self, value):
        self.state.settings.deconst_strategy = value

    @property
    def deconst_max_distance_t(self):
        return self.state.settings.deconst_max_distance_t

    @deconst_max_distance_t.setter
    def deconst_max_distance_t(self, value):
        self.state.settings.deconst_max_distance_t = value

    @property
    def deconst_full_tolerance(self):
        return self.state.settings.deconst_full_tolerance

    @deconst_full_tolerance.setter
    def deconst_full_tolerance(self, value):
        self.state.settings.deconst_full_tolerance = value

    @property
    def force_time_expansion(self):
        return self.state.settings.force_time_expansion

    @force_time_expansion.setter
    def force_time_expansion(self, value):
        self.state.settings.force_time_expansion = value

    @property
    def allow_weak_reliability(self):
        return self.state.settings.allow_weak_reliability

    @allow_weak_reliability.setter
    def allow_weak_reliability(self, value):
        self.state.settings.allow_weak_reliability = value

    @property
    def learn_block_minimum(self):
        return self.state.settings.learn_block_minimum

    @learn_block_minimum.setter
    def learn_block_minimum(self, value):
        self.state.settings.learn_block_minimum = value

    @property
    def min_reliability(self):
        return self.state.settings.min_reliability

    @min_reliability.setter
    def min_reliability(self, value):
        self.state.settings.min_reliability = value



#  C O N S T R U C T I V I S T__ __
#   _____ ____   ____   /  |/  // /
#  / ___// __ \ / __ \ / /|_/ // /
# / /__ / /_/ // / / // /  / // /___
# \___/ \____//_/ /_//_/  /_//_____/
#  M A C H I N E   L E A R N I N G
#
#
# A Project by
# Thomas Schmid | UNIVERSITÄT LEIPZIG
# www.constructivist.ml
#
# Code Author: Dmitrij Denisenko
# Licence: MIT


"""Defines the pandas adapter.

"""


import re
import pandas as pd


def convert_df_to_block(dataframe):
    """Create PandasBlock from given dataframe.

    Args:
        dataframe (pandas.DataFrame)

    Returns:
        PandasBlock

    """
    return PandasBlock(dataframe)


def build_block_from_rows(rows):
    """Create PandasBlock from a list of pandas.Series.

    Args:
        rows (list): Rows as pandas.Series.

    Returns:
        PandasBlock

    """
    df = pd.concat(rows, sort=False)
    return PandasBlock(df)


def build_new_learnblock(values, columns, index, origin):
    """Build PandasBlock from given values, columns and index.

    Args:
        values (numpy.ndarray): Data that forms the dataframe.
        columns (list): Names of the columns as str.
        index (list): Index to use.
        origin (tuple): Indicator where this learnblock comes from.

    Returns:
        PandasBlock

    """
    data_frame = pd.DataFrame(data=values, columns=columns, index=index)
    block = PandasBlock(data_frame, origin=origin)
    return block


class PandasBlock:
    """Responsible for encapsulating pandas.DataFrame.

    Args:
        dataframe (pandas.DataFrame): Contains the data.
        relationship (tuple):
            If not None, contains the learnblock relationship.
        subject (tuple): If not None, contains the learnblock subject.
        origin (tuple):
            If not None, contains the origin of the learnblock.
            Plain integers indicate that this learnblock is from a raw
            block. Aim values indicate that this learnlock was build
            form the labels from two models.

    """
    LAST_THREE_COLUMNS = 3
    TIME = "T"
    SUBJECT = "Sigma"
    PURPOSE = "Z"

    def __init__(self, dataframe, relationship=None, subject=None, origin=None):
        self.df = dataframe
        self._origin = origin
        self.relationship = relationship
        self.subject = subject
        # TODO (dmt): Check if you really need n_cluster.
        self.n_cluster = None

    @property
    def origin(self):
        return self._origin

    @origin.setter
    def origin(self, value):
        self._origin = value

    def __str__(self):
        return ("origin: <{}> subject: <{}> "
                "n_cluster: <{}> relationship: <{}>".format(self.origin,
                                                            self.subject,
                                                            self.n_cluster,
                                                            self.relationship))

    def __len__(self):
        return self.n_rows()

    def __getitem__(self, item):
        end = self.n_columns(effective=True)
        return self.df.iloc[item][:end]

    def shape(self):
        """Shape of the dataframe.

        Returns:
            tuple:
                Integers indicating row and column number.

        """
        return self.df.shape

    def row_as_df_via_index(self, index):
        """Get pandas.Series via given Index.

        Args:
            index (int): Row index.

        Returns:
            pandas.Series

        """
        return self.df.loc[[index]]

    def n_rows(self):
        """Number of rows.

        Returns:
           int

        """
        return self.df.shape[0]

    def n_columns(self, effective=False):
        """Number of columns.

        Args:
            effective (bool): Indicator if T, Sigma and Z should be taken
                into account.

        Returns:
            int:
                Number of columns.

        """
        if effective:
            return max(0, self.df.shape[1] - self.LAST_THREE_COLUMNS)
        else:
            return self.df.shape[1]

    def columns(self, effective=False):
        """Get the names of the columns.

        Args:
            effective (bool): Indicator if T, Sigma, Z should be taken into
                account.

        Returns:
            list:
                Names of columns.

        """
        if not effective:
            return list(self.df.columns)
        else:
            features = list(self.df.columns)
            features.remove(self.TIME)
            features.remove(self.PURPOSE)
            features.remove(self.SUBJECT)
            return features

    def memory_usage(self):
        """Memory usage of the dataframe.

        Returns:
            int

        """
        return self.df.memory_usage().sum()

    def min_timestamp(self):
        """Smallest time value within T feature.

        Returns:
            int

        """
        return min(self.df["T"])

    def max_timestamp(self):
        """Biggest time value withtin T feature.

        Returns:
            int

        """
        return max(self.df["T"])

    def indices(self):
        """Get the indices of pandas.DataFrame.

        Returns:
            list:
                Indices as integers.

        """
        return list(self.df.index)

    def reindex(self, indices):
        """Change the indices of dataframe to new indices.

        Args:
            indices (list): New indices.

        """
        assert len(indices) == self.df.shape[0]
        self.df.index = indices

    def has_nan(self):
        """Check if the dataframe has any NaN.

        Returns:
            bool
                True, if there is at least one NaN.

        """
        return bool(self.df.isna().any().any())

    def drop_rows(self, row):
        """Remove given row from the dataframe.

        Args:
            row (pandas.Series): Row to remove.

        """
        self.df.drop(row, inplace=True)

    def rename_columns(self, names):
        """Rename dataframe columns to the given names.

        Args:
            names (list): Names as strings.

        """
        names.extend(["T", "Sigma", "Z"])
        self.df.columns = names

    def rename_column(self, old_name, new_name):
        """Rename a single column.

        Args:
            old_name (str): Column to rename.
            new_name (str): New name.

        """
        self.df = self.df.rename(columns={old_name: new_name})

    def drop_columns(self, columns, by_index=False):
        """Remove columns from dataframe.

        Args:
            columns (list): String names or integer indices.
            by_index (bool): Indicates if columns should be interpreted as
                integers or strings.

        """
        if not by_index:
            self.df = self.df.drop(columns=columns)
        else:
            assert all([isinstance(column, int) for column in columns])
            names = [self.df.columns[index] for index in columns]
            self.df = self.df.drop(columns=names)

    def labeled(self):
        """Check if the block is alreay labeled.

        Returns:
            bool:
                True, if labeled.

        """
        values_empty = self.df["Z"].iloc[0].strip('""') == ''
        values_unique = self.df["Z"].nunique() == 1
        if values_empty and values_unique:
            return False

        elif not values_empty and values_unique:
            return True

        else:
            return False

    def reorder_columns(self, columns):
        """Reorder columns accoring the given order.

        Args:
            columns (list): Column names as strings.

        """
        self.df = self.df[columns]

    def as_numpy_array(self, effective=True):
        """Get the dataframe as numpy ndarray.

        Args:
            effective (bool): Indicator if columns T, Sigma, Z should be taken
                into account.

        Returns:
            pandas.ndarray

        """
        return self.df[self.columns(effective=effective)].values

    def set_column(self, name, values):
        """Assign column values to given values.

        If the given column does not exist, it will be created.

        Args:
            name (str): Column name.
            values (list): Column values.

        """
        if isinstance(values, list):
            assert len(values) == self.n_rows()
        self.df[name] = values

    def new_block_from(self, fundament, time_column=False, index=False):
        """Construct PandasBlock from either given time stamps or given
        index.

        Args:
            fundament (list): Timestamps or indices as intergers.
            time_column (bool): Indicates if fundament should be interpreted
                as a list of timestamps.
            index (bool): Indicates if fundament should be interpreted as
                as list of indices.

        Returns:
              PandasBlock.

        """
        if index:
            assert all([isinstance(index, int) for index in fundament])
            return PandasBlock(
                self.df.loc[fundament],
                self.relationship,
                self.subject,
                origin=self.origin)

        if time_column:
            assert all([isinstance(timestamp, int) for timestamp in fundament])
            return PandasBlock(
                self.df.loc[self.df["T"].isin(fundament)],
                self.relationship,
                self.subject,
                origin=self.origin)

    def get_column(self, name):
        """Get column values for the given column.

        Args:
            name (str): Column name.

        Returns:
            list

        """
        return self.df[name].to_list()

    def get_row(self, index):
        """Get row via index.

        Args:
            index (int): Row index.

        Returns:
            list

        """
        return self.df.iloc[index].to_list()

    def sort_after(self, column, ascending=True):
        """Sort column.

        Args:
            column (str): Column name.
            ascending (bool): Indicates sorting order.

        """
        self.df.sort_values(by=[column], ascending=ascending, inplace=True)

    def get_duplicated_pairs(self, column_a, column_b):
        """Find duplicated rows in the given columns.

        Args:
            column_a (str): Column name as str.
            column_b (str): Column name as str.

        Yield:
            tuple:
                Duplicated rows.

        """
        bool_series = self.df.duplicated(subset=[column_a, column_b])
        duplicates = self.df[bool_series]
        for i, j in zip(duplicates[column_a], duplicates[column_b]):
            yield i, j

    def construct_from_two_pairs(self, **kwargs):
        """Build PandasBlock from given T, Z or Sigma value.

        Args:
            **kwargs: Contains T, Z and Sigma value.

        Returns:
            PandasBlock

        """
        t, z, sigma = kwargs.get("T"), kwargs.get("Z"), kwargs.get("Sigma")

        if t is not None and z is not None:
            df = self.df.loc[(self.df["T"] == t) & (self.df["Z"] == z)]
        elif t is not None and sigma is not None:
            df = self.df.loc[(self.df["T"] == t) & (self.df["Sigma"] == sigma)]
        else:
            df = self.df.loc[(self.df["Z"] == z) & (self.df["Sigma"] == sigma)]

        return PandasBlock(df, self.relationship, origin=self.origin)

    def get_overlapping_on_timestamps(self, block):
        """Construct a new PandasBlock from overlapping timestamps.

        Args:
            block (PandasBlock): Block.

        Returns:
              PandasBlock

        """
        if len(block) < len(self):
            biggest = block
            block = self
        else:
            biggest = self
            block = block

        join = biggest.df.join(block.df.set_index("T"),
                               on="T",
                               lsuffix="_other",
                               how="inner")

        if any(join.duplicated(["T"])):
            join.drop_duplicates(subset="T", keep="first", inplace=True)
            # empty_df = pd.DataFrame()

        return PandasBlock(join, origin=self.origin, subject=self.subject)

    def same_features_fusion(self, block):
        """Build a new PandasBlock based on the feature intersection of
        two PandasBlocks.

        Args:
            block (PandasBlock): Learnblock.

        Returns:
            PandasBlock

        """
        df = pd.concat([self.df, block.df], sort=False, join="inner")
        origin = tuple(set(self.origin+block.origin))
        return PandasBlock(df, subject=self.subject, origin=origin)

    def constructed_from_models(self):
        """Check if this PandasBlock from build based on
        PragmaticMachineLearningModel.

        Returns:
            bool:
                True, if this block was build based on
                PragmaticMachineLearningModel.

        """
        regex = "[A-Z]\.\d+\.\d+"
        return any([re.match(regex, str(o)) for o in self.origin])



#  C O N S T R U C T I V I S T__ __
#   _____ ____   ____   /  |/  // /
#  / ___// __ \ / __ \ / /|_/ // /
# / /__ / /_/ // / / // /  / // /___
# \___/ \____//_/ /_//_/  /_//_____/
#  M A C H I N E   L E A R N I N G
#
#
# A Project by
# Thomas Schmid | UNIVERSITÄT LEIPZIG
# www.constructivist.ml
#
# Code Author: Dmitrij Denisenko
# Licence: MIT


"""Define the commands of the framework.

"""


from conML.usecases.usecase import Usecase
from conML.domain import (
    FeatureSelector,
    Reconstructor,
    Constructor,
    Deconstructor,
    ClusterFinder)


class ModelCreationMixin:
    def create_models(self, interface, algorithms):
        """Instantiated MachineLearningModels via the given interface.

        Args:
            interface (Callable): Adapter.
            algorithms (List): List of scikit-algorithms.

        Yield:
            MachineLearningModel

        """
        for abbreviation, algorithm in algorithms:
            model = interface(algorithm, abbreviation)
            yield model


class FeatureSelectionUsecase(Usecase):
    def process(self, request):
        """Process request for creating a FeatureSelector.

        Args:
            request (FeatureSelectionRequest): Namedtuple with all essential
                ingredients.

        Returns:
            FeatureSelector

        """
        return FeatureSelector(request.filter_method,
                               request.embedded_method,
                               request.settings)


class ReconstructionUsecase(Usecase, ModelCreationMixin):
    def process(self, request):
        """Process request for creating a Reconstructor.

        Args:
            request (ReconstructorRequest): Namedtuple with all essential
                ingredients.

        Returns:
            Reconstructor

        """
        models = list(self.create_models(request.interface, request.algorithms))
        return Reconstructor(request.mode, models, request.settings,
                             request.krippendorf)


class ConstructionUsecase(Usecase, ModelCreationMixin):
    def process(self, request):
        """Process request for creating a Constructor.

        Args:
            request (ConstructorRequest): Namedtuple with all essential
                ingredients.

        Returns:
            Constructor

        """
        models = list(self.create_models(request.interface, request.algorithms))
        constructor = Constructor(request.mode, models, request.settings)
        return constructor


class DeconstructionUsecase(Usecase):
    def process(self, request):
        """Process request for creating Deconstructor.

        Args:
            request (DeconstructionRequest): Namedtuple with all essential
                ingredients.

        Returns:
            Deconstructor

        """
        cluster_finder = ClusterFinder(request.density_estimator,
                                       request.block_processing_settings,
                                       request.ckmeans)
        return Deconstructor(request.general_settings,
                             request.block_processing_settings,
                             request.deconstruction_settings,
                             request.reconstructor,
                             request.build_new_learnblock,
                             cluster_finder)



#  C O N S T R U C T I V I S T__ __
#   _____ ____   ____   /  |/  // /
#  / ___// __ \ / __ \ / /|_/ // /
# / /__ / /_/ // / / // /  / // /___
# \___/ \____//_/ /_//_/  /_//_____/
#  M A C H I N E   L E A R N I N G
#
#
# A Project by
# Thomas Schmid | UNIVERSITÄT LEIPZIG
# www.constructivist.ml
#
# Code Author: Dmitrij Denisenko
# Licence: MIT


"""Defines adapters for machine machine algorithms of scikit-learn.

"""


from copy import deepcopy
from collections import Counter
import warnings

from sklearn.cluster import KMeans
from sklearn.cluster import Birch
from sklearn.cluster import SpectralClustering
from sklearn.cluster import DBSCAN
from sklearn.cluster import OPTICS
from sklearn.cluster import AffinityPropagation
from sklearn.cluster import AgglomerativeClustering
from sklearn.cluster import MeanShift
from scipy.stats import gaussian_kde
from numpy import random


CLUSTERING_SPECIFIC = {
    KMeans: True,
    DBSCAN: False,
    SpectralClustering: True,
    AgglomerativeClustering: True,
    AffinityPropagation: False,
    MeanShift: False,
    OPTICS: False,
    Birch: True,
}


class MachineLearningModel:
    """Base class for all machine learning models.

    """

    def train(self, data, *args, **kwargs):
        raise NotImplementedError


class FilterMethod(MachineLearningModel):
    """Responsible for encapsulating filter methods.

    Filter methods are used during the feature selection process.

    Args:
        model: Scikit-learn filter method.

    """

    def __init__(self, model):
        self.model = model

    def __str__(self):
        return "Filtermethod: {}".format(str(self.model))

    def __repr__(self):
        return "FilterMethod({})".format(repr(self.model))

    def train(self, data, *args, **kwargs):
        """Train the filter method on given data set.

        Args:
            data (PandasBlock): Learnblock.

        """
        labels = data.get_column("Z")
        self.model = self.model.fit(data.df.iloc[:, :-3], y=labels)

    def reduce(self, data):
        """Determine features that can be savely removed.

        Args:
            data (PandasBlock): Learnblock.

        Returns:
            set:
                Features that do not contribute much to the label.

        """
        supported_indices = set(self.model.get_support(indices=True))
        features_indices = {i for i in range(data.n_columns(effective=True))}
        return features_indices.difference(supported_indices)


class EmbeddedMethod(MachineLearningModel):
    """Resposible for encapsulating embedded methods.

    Args:
        model: Scikit-learn embedded method.

    """
    _ONE_HUNDRET_PERCENT = 100

    def __init__(self, model):
        self.model = model

    def __str__(self):
        return "EmbeddedMethod: {}".format(str(self.model))

    def __repr__(self):
        return "EmbeddedMethod({})".format(repr(self.model))

    def train(self, data, *args, **kwargs):
        """Train the embedded method on given data set.

        Args:
            data (PandasBlock): Learnblock.

        """
        labels = data.get_column("Z")
        self.model = self.model.fit(data.df.iloc[:, :-3], labels)

    def reduce(self, data):
        """Determine features that can be savely removed from the given
        learnblock.

        Args:
            data (PandasBlock):  Learnblock.

        Returns:
            set:
                Features that do not contribute much to the label.

        """
        supported_indices = set(self.model.get_support(indices=True))
        features_indices = {i for i in range(data.n_columns(effective=True))}
        return features_indices.difference(supported_indices)


class ConstructionClusteringMLModel(MachineLearningModel):
    """Responsible for encapsulating unsupervised machine learning
    models from scitkit-learn.

    Args:
        model: Unsupervised machine learning model from scikit-learn.
        abbreviation (str): Subject identifier.

    """
    def __init__(self, model, abbreviation):
        self.model = model
        self.abbreviation = abbreviation
        self._cluster = None

    def __str__(self):
        return "\n".join([
                "{}: {}".format("Model", str(self.model)),
                "{}: {}".format("Abbreviation", str(self.abbreviation)),
           ]
        )

    def __repr__(self):
        return ",".join([
                "{}={}".format("Model", repr(self.model)),
                "{}={}".format("Abbreviation", str(self.abbreviation)),
                "{}={}".format("Cluster", str(self._cluster))
            ]
        )

    def new_model_depending_on_cluster(self, n_cluster):
        """Factory method for creating ConstructionClusteringMLModel
        with a specific cluster number.

        Args:
            n_cluster (int): Number of cluster that the unsupervised
                machine learning model should find.

        Returns:
            ConstructionClusteringMLModel

        """
        new_model = deepcopy(self)
        new_model._cluster = n_cluster
        new_model.model.n_clusters = n_cluster
        return new_model

    def new_model(self):
        """Create a deep copy of itself.

        Returns:
              ConstructionClusteringMLModel

        """
        return deepcopy(self)

    def get_labels(self):
        """Return found labels

        Returns:
            numpy.ndarray

        """
        return self.model.labels_

    @property
    def cluster_specific(self):
        """Check if the used unsupervised machine learning model needs
        a cluster specification.

        Returns:
            bool:
                True, if the number of cluster that needs to be found
                should be specified by the user.

        """
        return CLUSTERING_SPECIFIC[type(self.model)]

    @property
    def found_cluster(self):
        """Number of found clusters.

        Returns:
            int

        """
        return len(set(self.model.labels_))

    @property
    def cluster(self):
        return self._cluster

    @cluster.setter
    def cluster(self, value):
        self.model.n_clusters = value
        self._cluster = value

    @property
    def cluster_sizes(self):
        return Counter(self.model.labels_)

    def train(self, data, *args, **kwargs):
        """Train the unsupervised machine learning model on the given learnblock.

        Args:
            data (PandasBlock): Learnblock.

        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.model.fit(data)


class ReconstructionConceptualMLModel(MachineLearningModel):
    """Responsible for encapsulating supervised machine learning
    models from scitkit-learn.

    Args:
        model: Supervised machine learning model from scikit-learn.
        abbreviation (str): Subject identifier.

    Attributes:
        accuracy (float): Prediction accuracy on test data.

    """

    def __init__(self, model, abbreviation):
        self.model = model
        self.subject = abbreviation
        self.accuracy = None

    def train(self, data, *args, **kwargs):
        """Train the supervised machine learning model on the given data set.

        Args:
            data (numpy.ndarray): Training data.

        Returns:
            ReconstructionConceptualMLModel

        """
        labels = args[0]
        model = self.model.fit(data, labels)
        recon_model = deepcopy(self)
        copy_model = deepcopy(model)
        recon_model.model = copy_model
        return recon_model

    def score(self, data, labels):
        """Calculate the accuracy on the given data set.

        Args:
            data (numpy.ndarray): Data set.
            labels (list): True labels for the data set.

        Returns:
            float:
                Accuracy.

        """
        return round(self.model.score(data, labels), 3)

    def predict(self, data):
        """Predict the labels for the given data set.

        Args:
            data (numpy.ndarray): Data set.

        Returns:
            list:
                Predicted labels for the given data set.

        """
        return [i for i in self.model.predict(data)]


class KernelDensityEstimator:

    def __init__(self):
        """Responsible for encapsulating  kernel density estimation
        algorithms from scikit-learn.

        """
        self.model = gaussian_kde

    def density(self, data):
        """Calculate the density of the given data.

        Args:
            data (list): List with time stamps.

        Returns:
            list:
                List of floats representing the denisty values.

        """
        return gaussian_kde(data)(data)



#  C O N S T R U C T I V I S T__ __
#   _____ ____   ____   /  |/  // /
#  / ___// __ \ / __ \ / /|_/ // /
# / /__ / /_/ // / / // /  / // /___
# \___/ \____//_/ /_//_/  /_//_____/
#  M A C H I N E   L E A R N I N G
#
#
# A Project by
# Thomas Schmid | UNIVERSITÄT LEIPZIG
# www.constructivist.ml
#
# Code Author: Dmitrij Denisenko
# Licence: MIT


"""Defines the block drawing, learnlbock identification and cluster search
within one dimensional list.

"""


from itertools import zip_longest

from conML.shared import logger
from conML.shared.logger import LearnblockIdentificationInfo


__all__ = (
    "LearnblockIdentifier",
    "ClusterFinder",
    "Source"
)


class Source:
    """Responsible of preprocessing the given pandas dataframes.

    Converts the the given pandas dataframe to an PandasBlock object and
    saves it into knowledge database.

    Args:
        knowledge (KnowledgeDatabase): Found blocks written to this database.
        df_converter (Callable): Convert pandas dataframe to an PandasBlock.

    """

    def __init__(self, knowledge, df_converter):
        self.knowledge = knowledge
        self.df_converter = df_converter

    def process(self, df):
        """Converts the given dataframe into PandasBlock and logs this
        process.

        Args:
            df (pandas.Dataframe): User provided data as dataframe.

        Returns:
            tuple:
                Logging information and the created PandasBlock.

        """
        pandas_block = self.convert_to_pandas_block(df)
        info = logger.BlockProcessInfo(
            pandas_block.n_rows(),
            pandas_block.n_columns()
        )
        return info, pandas_block

    def convert_to_pandas_block(self, df):
        """Converts the given dataframe into PandasBlock.

        Args:
            df (pandas.Dataframe): User provided data as dataframe.

        Returns:
            PandasBlock

        """
        return self.df_converter(df)

    def append_to_knowledge(self, pandas_block):
        """Inserts the a PandasBlock object into knowledge database.

        Args:
            pandas_block (PandasBlock): Valid instantiated PandasBlock.

        """
        n_block = self.knowledge.extend(pandas_block)
        pandas_block.origin = (n_block, )


class LearnblockIdentifier:
    """Responsible for identification of valid learnblocks.

    Args:
        settings (BlockProcessingSettings): Defines the behaviour of the
            learnblock identification process.
        density_estimator (MachineLearningModel): Estimates the density
            of timestamps.
        ckmeans (Callable): Performs the optimal kmeans on a 1-dimensional
            list.
    """
    def __init__(self, settings, density_estimator, ckmeans):
        self.settings = settings
        self.cluster_finder = ClusterFinder(
            density_estimator, settings, ckmeans
        )

    @classmethod
    def _column_pairs(cls):
        """Yiels the possible relationships for valid learnblocks.

        Yields:
            tuple

        """
        yield "Sigma", "Z"
        # yield "T", "Sigma"
        # yield "T", "Z"

    def identify(self, block):
        """Identify the biggest learnbock within the given PandasBlock.

        Args:
            block (PandasBlock): Instantiated PandasBlock.

        Returns:
            PandasBlock:
                Subset of the given PandasBlock.

        """
        biggest_learn_block = None
        biggest_block_size = 0

        for pair in LearnblockIdentifier._column_pairs():
            for possible_learnblock in self._identify_relatives(block, *pair):
                if self._is_learn_block(possible_learnblock.n_rows()):
                    if possible_learnblock.n_rows() > biggest_block_size:
                        biggest_learn_block = possible_learnblock
                        biggest_learn_block.relationship = pair

        if biggest_learn_block is None:
            return (
                logger.LearnblockIdentificationInfo(
                    LearnblockIdentificationInfo.State.DISCARDED
                ), None, self.get_halde(biggest_learn_block, block))

        else:
            return (
                logger.LearnblockIdentificationInfo(
                    LearnblockIdentificationInfo.State.IDENTIFIED,
                    biggest_learn_block.n_rows(),
                    biggest_learn_block.n_columns(),
                    biggest_learn_block.relationship
                ),
                biggest_learn_block,
                self.get_halde(biggest_learn_block, block)
            )

    def _is_learn_block(self, block_size):
        """Check if the given block size is enough to be classified as a
        learnblock.

        Args:
            block_size (int): Number of samples in a block.

        Returns:
            bool:
                Returns True if the size is enough for a learnblock else False.

        """
        return block_size >= self.settings.learn_block_minimum

    def _identify_relatives(self, block, *args):
        """Find relatives within the given block. And used them to coconstruct
        a new block.

        Args:
            block (PandasBlock): Instantiated PandasBlock.
            *args: Relationship to be used.

        Yiels:
            PandasBlock:
                Constructed from relatives.

        """
        already_seen = set()
        for value_pair in block.get_duplicated_pairs(args[0], args[1]):
            if value_pair not in already_seen:
                already_seen.add(value_pair)

                kw = {args[0]: value_pair[0], args[1]: value_pair[1]}
                if args[0] == "Sigma" and args[1] == "Z":
                    for lb_cand in self._get_sigma_zeta_relatives(block, **kw):
                        yield lb_cand
                else:
                    yield block.construct_from_two_pairs(**kw)

    def _get_sigma_zeta_relatives(self, block, **kw):
        """Find Sigma Z relatives within the given block and used them to
        construct a new block.

        Args:
            block (PandasBlock): Instantiated PandasBlock.
            **kw: Sigma Z relative.

        Yields:
            PandasBlock:
                Possible learnblock with Sigma Z relatives.

        """
        relatives = block.construct_from_two_pairs(**kw)
        time_column = relatives.get_column("T")
        clusters = self.cluster_finder.find(time_column)
        for time_values in clusters:
            yield relatives.new_block_from(time_values, time_column=True)

    def get_halde(self, learnblock, block):
        """Get block subset which is not part of the learnblock.

        Args:
            learnblock (PandasBlock): Biggest valid found learnblock.
            block (PandasBlock): User provided block.

        Returns:
            pandas.Dataframe:
                Subset which is not part of learnblock.

        """
        if learnblock is None:
            return block.df
        else:
            origin_block = set(block.indices())
            lb_index = set(learnblock.indices())
            halde_indices = list(origin_block.difference(lb_index))
            return block.new_block_from(halde_indices, index=True).df


class ClusterFinder:
    """Responsible for finding the optimal number of clusters within
    timestamps.

    Args:
        density_estimator (MachineLearningModel):
            Estimates the density of the timestamps.
        settings (BlockProcessingSettings):
            Defines the behavior of the cluster finding process.
        ckmeans (Callable):
            Interface to the ckmeans algorithm.

    """
    def __init__(self, density_estimator, settings, ckmeans):
        self.density_estimator = density_estimator
        self.settings = settings
        self.ckmeans = ckmeans

    def find(self, time_column):
        """Find the optimal number of cluster in the given timestamps.

        Args:
            time_column (list): List with timestamps as integers.

        Returns:
            list:
                List timestamp clusters.

        """
        density = self.density_estimator.density(time_column)
        changes = self.count_threshold_changes(density)
        n_changes = self.count_changes(changes)
        return self.ckmeans(time_column, n_changes)

    def find_valid_learnblock_cluster(self, time_column):
        """Find the optimal number of clusters in the given timestamps,
        but every cluster should contain enough timestamps to be classified
        as a valid learnblock.

        Args:
            time_column (list): List with timestamps as integers.

        Yields:
            list:
                Timestamps that form a valid learnblock.

        """
        density = self.density_estimator.density(time_column)
        changes = self.count_threshold_changes(density)
        n_changes = self.count_changes(changes)
        cluster = self.ckmeans(time_column, n_changes)
        for c in cluster:
            if (len(c)/len(time_column)) > self.settings.min_category_size:
                yield c

    def count_threshold_changes(self, density):
        """Counts how many times the density in a given time stamp
        exceeds the threshold value.

        Args:
            density (Iterable): Density estimation over the timestamps.

        Returns:
            List:
                List of booleans. True indicates the density exceeded the
                threshold otherwise False.

        """
        changes = list()
        max_dens = max(density)
        for index, dens in enumerate(density):
            if dens > max_dens*self.settings.sigma_zeta_cutoff:
                changes.append(True)
            else:
                changes.append(False)
        return changes

    def count_changes(self, changes):
        """Count the number of clusters to be found by ckmeans.

        Consecutive True, False or False, True indicates the beginning or
        end of a cluster.


        Args:
            changes (list): List of booleans.

        Returns:
            int:
                Number of clusters.

        """
        count_changes = 0
        for i, j in zip_longest(changes[0::1], changes[1::1]):
            if (i, j) == (True, False) or (i, j) == (False, True):
                count_changes += 1

        if count_changes == 0:
            return 1

        return count_changes+1



#  C O N S T R U C T I V I S T__ __
#   _____ ____   ____   /  |/  // /
#  / ___// __ \ / __ \ / /|_/ // /
# / /__ / /_/ // / / // /  / // /___
# \___/ \____//_/ /_//_/  /_//_____/
#  M A C H I N E   L E A R N I N G
#
#
# A Project by
# Thomas Schmid | UNIVERSITÄT LEIPZIG
# www.constructivist.ml
#
# Code Author: Dmitrij Denisenko
# Licence: MIT


from abc import abstractmethod, ABC


class Usecase(ABC):
    """Base class for all use cases.

    """
    def execute(self, request):
        """Execute the given requests.

        Args:
            request (namedtuple): Request with important parameters.

        Returns:
            Any

        """
        try:
            return self.process(request)

        except Exception as error:
            print(error.with_traceback())
            raise

    @abstractmethod
    def process(self, *args, **kwargs):
        """Every use case should implent their logic within this method. """
        raise NotImplementedError



#  C O N S T R U C T I V I S T__ __
#   _____ ____   ____   /  |/  // /
#  / ___// __ \ / __ \ / /|_/ // /
# / /__ / /_/ // / / // /  / // /___
# \___/ \____//_/ /_//_/  /_//_____/
#  M A C H I N E   L E A R N I N G
#
#
# A Project by
# Thomas Schmid | UNIVERSITÄT LEIPZIG
# www.constructivist.ml
#
# Code Author: Dmitrij Denisenko
# Licence: MIT


"""Define the logging mechanism.


"""


import json
from os.path import join
from enum import Enum


class LogMeMixin:
    def log_me(self, logger, stdout, json_container):
        """Log the information gathered during the knowledge search process.

        Args:
            logger (logging.Logger):
            stdout (bool): Indicator if it should be logged to console.
            json_container (JsonContainer): Container for the logging
                information in JSON-Format.

        """
        json_container.log(self)

        if stdout:
            logger.protocol(self.banner())
            logger.protocol(str(self))


class BlockProcessInfo(LogMeMixin):
    """Holds logging information of the block pulling process.

    Args:
        n_rows (int): Number of rows the block contains.
        n_columns (int): Number of coloumns the block contains.

    Attributes:
         level (int): Level on which the block is pulled.

    """

    def __init__(self, n_rows, n_columns):
        self.level = 1
        self.n_rows = n_rows
        self.n_columns = n_columns

    def __str__(self):
        return "\n".join([
            "{: <20}{}".format("Rows", str(self.n_rows)),
            "{: <20}{}".format("Columns", str(self.n_columns))
        ])

    def banner(self):
        """Return banner that indicates the beginning of the learnblock
        identification.

        Returns:
            str

        """
        return "{:=^100}".format("BLOCKPROCESSING")


class LearnblockIdentificationInfo(LogMeMixin):
    """Holds logging information of the learnblock identification.

    Args:
        state (LearnblockIdentificationInfo.State):
        n_rows (int): Number of rows the learnblock has.
        n_columns (int): Number of columns the learnblock has.
        relationship (tuple): Which relationship was found to build
            the learnblock.

    Attributes:
         state (LearnblockIdentificationInfo.State): Indicator if the
            learnblock identification was successfull or not.

    """

    class State(Enum):
        IDENTIFIED = 1
        DISCARDED = 2

    def __init__(self, state, n_rows=None, n_columns=None, relationship=None):
        self.state = state
        self.n_rows = n_rows
        self.n_columns = n_columns
        self.relationship = relationship

    def __str__(self):
        return "\n".join([
            "{: <20}{}".format("State", str(self.state)),
            "{: <20}{}".format("Relationship", str(self.relationship))
        ])

    def banner(self):
        """Return banner that indicates the beginning of the learnblock
        identification.

        Returns:
            str

        """
        return "{:=^100}".format("LEARNBLOCK")


class HighLevelLearnblockInfo(LogMeMixin):
    """Holds logging information about the learnblock pushed on the higher
    level.

    Args:
        lb (PandasBlock): Learnblock.
        level (int): Level on which the learnblock is pushed.

    Attributes:
         n_rows (int): Number of rows the learnbock has.
         n_columns (int): Number of columns the learnblock has.
         origin (tuple): Indicates who was the learnblock builder.

    """

    def __init__(self, lb, level):
        self.level = level
        self.n_rows = lb.n_rows()
        self.n_columns = lb.n_columns()
        self.origin = lb.origin

    def __str__(self):
        return "\n".join([
            "{: <20}{}".format("Level", str(self.level)),
            "{: <20}{}".format("Rows", str(self.n_rows)),
            "{: <20}{}".format("Columns", str(self.n_columns))
        ])

    def banner(self):
        """Return banner that indicates the beginning of the learnblock.

        Returns:
            str

        """
        return "{:=^100}".format("LEARNBLOCK")


class ConstructionInfo(LogMeMixin):
    """Holds logging information about the construction process.

    Args:
        state (ConstructionInfo.State): Indicates if the construction
            process was successfull or not.
        learnblock (PandasBlock): Constructed learnblock.
        subject (tuple): Name of the unsupervised machine learning model.
        n_cluster (int): Number of cluster the constructed learnblock has.

    Attributes:
        rows (int): Number of rows the learnblock has.
        columns (int): Number columns the learnblock has.
        min_t (int): Minimum timestamp the T features has.
        max_t (int): Maximum timestamp the T features has.
        origin (tuple): Indicates who was the learnblock builder.
        relationship (tuple): Indicates which relationship if any was
            used to buld the learnblock.

    """

    class State(Enum):
        DISCARDED = 1
        CONSTRUCTED = 2
        PASSED = 3

    def __init__(self, state, learnblock=None, subject="", n_cluster=""):
        self.state = state
        self.subject = subject
        self.n_cluster = n_cluster

        # learnblock information
        if learnblock:
            self.rows = str(learnblock.n_rows())
            self.columns = str(learnblock.n_columns())
            self.min_t = str(learnblock.min_timestamp())
            self.max_t = str(learnblock.max_timestamp())
            self.origin = str(learnblock.origin)
            self.relationship = str(learnblock.relationship)
            self.lb_subject = learnblock.subject if learnblock.subject else ""

    def __str__(self):
        if self.state == self.State.PASSED:
            return "\n".join([
                    "{: <20}{}".format("State:", "PASSED"),
                    self.learnblock_info()
                ]
            )

        elif self.state == self.State.DISCARDED:
            return "\n".join([
                "{: <20}{}".format("State:", "DISCARDED"),
                "{: <20}{}".format("Subject:", self.subject),
                "{: <20}{}".format("Cluster:", self.n_cluster)
                ]
            )

        elif self.state == self.State.CONSTRUCTED:
            return "\n".join([
                    "{: <20}{}".format("State:", "CONSTRUCTED"),
                    self.learnblock_info()
                ]
            )

    def learnblock_info(self):
        return "\n".join([
            "{: <20}{}".format("Rows", self.rows),
            "{: <20}{}".format("Columns", self.columns),
            "{: <20}{}".format("Cluster", self.n_cluster),
            "{: <20}{}".format("MinT", self.min_t),
            "{: <20}{}".format("MaxT", self.max_t),
            "{: <20}{}".format("Origin", self.origin),
            "{: <20}{}".format("Relationship", self.relationship),
            "{: <20}{}".format("Subject", self.lb_subject)
            ]
        )

    def banner(self):
        """Return banner that indicates the beginning of the construction.

        Returns:
            str

        """
        return "{:=^100}".format("CONSTRUCTION")


class SelectionInfo(LogMeMixin):
    """Hold logging information of the feature selection process.

    Args:
        state (SelectionInfo.State): Indicates the state of the feature
            selection process.
        learnblock (PandasBlock): Learnblock.

    Attributes:
         n_rows (int): Number of learnblock rows.
         n_columns (int): Number of learnblock columns.

    """
    class State(Enum):
        REDUCIBLE = 1
        NON_REDUCIBLE = 2

    def __init__(self, state, learnblock):
        self.state = state
        self.n_rows = learnblock.n_rows()
        self.n_columns = learnblock.n_columns(effective=True)

    def __str__(self):
        return "\n".join([
            "{: <20}{}".format("State", str(self.state)),
            "{: <20}{}".format("Rows", str(self.n_rows)),
            "{: <20}{}".format("Columns", str(self.n_columns))
            ]
        )

    def banner(self):
        """Return banner that indicates the beginning of the feature selection
        process.

        Returns:
            str

        """
        return "{:=^100}".format("SELECTION")


class ReconstructionInfo(LogMeMixin):
    """Holds infomration of the reconstruction process.

    Args:
        state (ReconstructionInfo.State): Indicates if the reconstruction
            process was successfull.
        uid (str): Unique identifier of PragmaticMachineLearningModel.
        aim (str): Aim in the context of conML.
        accuracy (float): Accuracy of the supersived machine learnign model.
        subject (tuple): Abbreviations of the supervised machine learning
            model.s
        reliability (float): Inter-reliability between supervised machine
            learning models.

    """

    class State(Enum):
        DISCARDED = 1
        RECONSTRUCTED = 2

    def __init__(self, state=None, uid=None, aim=None, accuracy=None,
                 subject=None, reliability=0.0):
        self.state = state
        self.uid = uid
        self.aim = aim
        self.accuracy = accuracy
        self.subject = subject
        self.reliability = reliability

    def __str__(self):
        if self.state == ReconstructionInfo.State.RECONSTRUCTED:
            return "\n".join([
                    "{: <20}{}".format("State", str(self.state)),
                    "{: <20}{}".format("Pragmatic:", str(self.uid)),
                    "{: <20}{}".format("Subject:", str(self.subject)),
                    "{: <20}{}".format("Accuracy:", str(self.accuracy)),
                    "{: <20}{}".format("Reliability:", str(self.reliability)),
                    "\n"
                ]
            )

        elif self.state == ReconstructionInfo.State.DISCARDED:
            return "\n".join([
                "{: <20}{}".format("State", str(self.state)),
                "{: <20}{}".format("Subject:", str(self.subject)),
                "{: <20}{}".format("Accuracy:", str(self.accuracy)),
                "\n"
                ]
            )

    def banner(self):
        """Return banner that indicates the beginning reconstruction process.

        Returns:
            str

        """
        return "{:=^100}".format("RECONSTRUCTION")


class WinnerSelectionInfo(LogMeMixin):
    """Holds the information about the winner selection process.

    Args:
        state (WinnerSelectionInfo.State): Indicates the status of the
            winner selection process.
        uid (str): Unique identifier of the PragmaticMachineLearningModel.
        subject (tuple): Abbreviations of supervised machine learning
            models.
        accuracy (float): Accuracy of supervised machine learnign model.
        reliability (float): Inter-reliability between supervised machine
            leanring models.

    """

    class State(Enum):
        SELECTED = 1
        FAILED = 2

    def __init__(self, state, uid=None, subject=None, accuracy=None,
                 reliability=None):
        self.state = state
        self.uid = uid
        self.subject = subject
        self.accuracy = accuracy
        self.reliability = reliability

    def __str__(self):
        state_info = "{: <20}{}".format("State", str(self.state))
        if self.state == self.State.SELECTED:
            winner_info = "\n".join([
                "{: <20}{}".format("Reconstructed:", str(self.uid)),
                "{: <20}{: <15}{}".format("", "Subject:", str(self.subject)),
                "{: <20}{: <15}{}".format("", "Accuracy:", str(self.accuracy)),
                "{: <20}{: <15}{}".format("", "Reliability:",
                                          str(self.reliability))
                ]
            )
        else:
            winner_info = ""

        return state_info+"\n"+winner_info

    def banner(self):
        """Return the banner that indicates the beginning of the winner
        selection process.

        Returns:
            str

        """
        return "{:=^100}".format("WINNER SELECTION")


class DeconstructionInfo(LogMeMixin):
    """Holds the informatino about the deconstruction process.

    Args:
        state (DeconstructionInfo.State): Indicates if the deconstruction
            process was successful or not.
        relationship (tuple): Type of deconstruction.
        relative (PragmaticMachineLearningModel): Relative of the
            deconstructed.
        pragmatic (PragmaticMachineLearningModel): The deconstructed..
        reconstructed_info (list): Contains reconstructed
            PragmaticMachineLearningModels.
        deleted (list): Contains deleted PragmaticMachineLearningModel.
        inserted (list): Contains inserted into database
            PragmaticMachineLearningModels.
        new_learnblock (PandasBlock): Learnblock pushed on the next
            level after successfully TSigma deconstruction.

    """
    class State(Enum):
        DECONSTRUCTED = 1
        FAILED = 2

    def __init__(self, state=None, relationship=None, relative=None,
                 pragmatic=None, reconstructed_info=None, deleted=None,
                 inserted=None, new_learnblock=None):
        self.state = state
        self.relationship = relationship
        self.relative = relative
        self.pragmatic = pragmatic
        self.reconstructed_info = reconstructed_info if reconstructed_info else []
        self.deleted = deleted if deleted else []
        self.inserted = inserted if inserted else []
        self.new_learnblock = new_learnblock

    def __str__(self):
        msg = ""
        if self.relationship:
            msg = "\n".join([
                    msg,
                    "{: <20}{}".format("State:", str(self.state)),
                    "{:+^100}".format(str(self.relationship)),
                    "{: <20}{}".format("Pragmatic:", str(self.pragmatic.uid)),
                    "{: <20}{}".format("Relative:", str(self.relative.uid)),
                    ]
                )

        if self.relationship == ("Sigma", "T"):
            return "\n".join([
                msg,
                "{: <20}[{:^5}{:^5}]".format(
                    "Dimension: ", str(self.new_learnblock.n_rows()),
                                   str(self.new_learnblock.n_columns())),
                "{: <20}{}".format("Min T: ",
                                   self.new_learnblock.min_timestamp()),
                "{: <20}{}".format("Max T: ",
                                   self.new_learnblock.max_timestamp()),
                "{: <20}{}".format("Origin: ",
                                   str(self.new_learnblock.origin)),
                ]
            )

        return "\n".join([
                msg,
                self.reconstructed_info_str(),
                "{: <20}{}".format(
                    "Inserted:", [", ".join([str(i.uid) for i in self.inserted
                                             if i is not None])]),
                "{: <20}{}".format(
                    "Deleted:", [", ".join([str(i.uid) for i in self.deleted
                                            if i is not None])])
            ]
        )

    def reconstructed_info_str(self):
        """Convert the PragmaticMachineLearningModels that were reconstructed
        during a deconstruction process into a string representation.

        Returns:
            str

        """
        msg = ""
        for m in self.reconstructed_info:
            msg = "\n".join([
                    msg,
                    "{: <20}{}".format("Reconstructed:", str(m.uid)),
                    "{: <20}{: <15} {}".format("", "Subject:", str(m.subject)),
                    "{: <20}{: <15} {}".format("", "Accuracy:", str(m.accuracy)),
                    "{: <20}{: <15} {}".format("", "Reliability:", str(m.reliability))
                ]
            )
        return msg

    def banner(self):
        """Return the banner that indicates the beginning of deconstruction
        process.

        Returns:
            str

        """
        return "{:=^100}".format("DECONSTRUCTION")


class JsonContainer:
    """Responsible for holding logging information about the knowledge
    search process in a JSON-FORMAT.

    Attributes:
        journal (dict): Container for the logging information.
        current_n_block (int): Currently processed block number.
        current_level (int): Level on which the currently learnblock
            is processed.
        current_lb (int): Currently processed learnblock.
        current_construction_subject (tuple): Currently used
            unsupervised machine learning model.
        dispatch_table (dict): Redirect the log object to the method
            responsible for storing his information into the journal.

    """
    def __init__(self):
        self.journal = {}
        self.current_n_block = 0
        self.current_level = 1
        self.current_lb = 0
        self.current_construction_subject = None

        self.dispatch_table = {
            BlockProcessInfo: self.log_process_block,
            LearnblockIdentificationInfo: self.log_lb_identification,
            HighLevelLearnblockInfo: self.log_higher_level_blocks,
            ConstructionInfo: self.log_construction,
            SelectionInfo: self.log_feature_selection,
            ReconstructionInfo: self.log_reconstruction,
            WinnerSelectionInfo: self.log_winner,
            DeconstructionInfo: self.log_deconstruction

        }

    def write_down(self, log_destination):
        """Write down the journal into the given destination.

        Args:
            log_destination (str): Logging destination.

        """
        file = join(log_destination, "log.json")
        with open(file, "w") as fp:
            json.dump(self.journal, fp)

    def log(self, info):
        """Call the responsible method for logging the given log object.

        Args:
            info: Log object.

        """
        self.dispatch_table[type(info)](info)

    def log_process_block(self, info):
        """Store the block processing logs into journal.

        Args:
            info (BlockProcessInfo): Holds block processing information.

        """
        self.current_n_block += 1
        self.journal[self.current_n_block] = {}
        self.journal[self.current_n_block]["BLOCK"] = {
                "rows": str(info.n_rows),
                "columns": str(info.n_columns),
                "LEVEL": {}
        }

    def log_lb_identification(self, info):
        """Store the learnblock identification logs into journal.

        Args:
            info (LearnblockIdentificationInfo): Holds learnblock identification
                information.

        """
        self.current_level = info.level
        level = self.journal[self.current_n_block]["BLOCK"]["LEVEL"]
        current_level = level[self.current_level] = {}

        current_level["LEARNBLOCKS"] = []
        current_level["LEARNBLOCKS"].append(
            {
                "rows": str(info.n_rows),
                "columns": str(info.n_columns),
                "relationship": str(info.relationship),
                "CONSTRUCTION": {},
                "WINNER": {},
                "DECONSTRUCTION": {
                    "COMPLETE": [],
                    "T SIGMA": [],
                    "Z SIGMA": []
                }
            }
        )
        self.current_lb = len(current_level["LEARNBLOCKS"]) - 1

    def log_higher_level_blocks(self, info):
        """Store learnblock logs that is considered for higher levels into
        journal.

        Args:
            info (HighLevelLearnblockInfo): Holds learnblock identification
                information.

        """
        self.current_level = info.level
        level = self.journal[self.current_n_block]["BLOCK"]["LEVEL"]

        if self.current_level not in level.keys():
            level[self.current_level] = {}

        current_level = level[self.current_level]
        if "LEARNBLOCKS" not in current_level.keys():
            current_level["LEARNBLOCKS"] = []

        current_level["LEARNBLOCKS"].append(
            {
                "rows": str(info.n_rows),
                "columns": str(info.n_columns),
                "origin": str(info.origin),
                "CONSTRUCTION": {},
                "WINNER": {},
                "DECONSTRUCTION": {
                    "COMPLETE": [],
                    "T SIGMA": [],
                    "Z SIGMA": []
                }
            }
        )
        self.current_lb = len(current_level["LEARNBLOCKS"]) - 1

    def log_construction(self, info):
        """Store construction log into journal.

        Args:
            info (ConstructionInfo): Construction logs.

        """
        self.current_construction_subject = str(info.subject)+str(info.n_cluster)
        key = (self.journal[self.current_n_block]
                           ["BLOCK"]
                           ["LEVEL"]
                           [self.current_level]
                           ["LEARNBLOCKS"]
                           [self.current_lb]
                           ["CONSTRUCTION"])
        key[self.current_construction_subject] = {"status": str(info.state)}

    def log_feature_selection(self, info):
        """Store features selection log into journal.

        Args:
            info (SelectionInfo): Feature selection logs.

        """
        key = (self.journal[self.current_n_block]
                           ["BLOCK"]
                           ["LEVEL"]
                           [self.current_level]
                           ["LEARNBLOCKS"]
                           [self.current_lb]
                           ["CONSTRUCTION"]
                           [self.current_construction_subject])

        key["SELECTION"] = {
            "rows": str(info.n_rows),
            "columns": str(info.n_columns)
        }

    def log_reconstruction(self, info):
        """Store reconstruction log into journal.

        Args:
            info (ReconstructionInfo): Reconstruction logs.

        """
        key = (self.journal[self.current_n_block]
                           ["BLOCK"]
                           ["LEVEL"]
                           [self.current_level]
                           ["LEARNBLOCKS"]
                           [self.current_lb]
                           ["CONSTRUCTION"]
                           [self.current_construction_subject])

        if "RECONSTRUCTION" not in key.keys():
            key["RECONSTRUCTION"] = []

        else:
            if info.state == ReconstructionInfo.State.RECONSTRUCTED:
                key["RECONSTRUCTION"].append(
                    {
                        "state": str(info.state),
                        "uid": str(info.uid),
                        "subject": str(info.subject),
                        "aim": str(info.aim),
                        "accuracy": str(info.accuracy),
                        "reliability": str(info.reliability)
                    }
                )

            else:
                key["RECONSTRUCTION"].append(
                    {
                        "state": str(info.state),
                        "subject": str(info.subject),
                        "accuracy": str(info.accuracy),
                        "reliability": str(info.reliability)
                    }
                )

    def log_winner(self, info):
        """Store winner selection log into journal.

        Args:
            info (WinnerSelectionInfo): Winner selection log.

        """
        key = (self.journal[self.current_n_block]
                           ["BLOCK"]
                           ["LEVEL"]
                           [self.current_level]
                           ["LEARNBLOCKS"]
                           [self.current_lb])

        if info.state == WinnerSelectionInfo.State.SELECTED:
            key["WINNER"] = {
                "state": str(info.state),
                "uid": str(info.uid),
                "subject": str(info.subject),
                "accuracy": str(info.accuracy),
                "reliability": str(info.reliability),
            }

        else:
            key["WINNER"] = {
                "state": str(info.state),
            }

    def log_deconstruction(self, info):
        """Store deconstruction log into journal.

        Args:
            info (DeconstructionInfo): Deconstruction log.

        """
        key = (self.journal[self.current_n_block]
                           ["BLOCK"]
                           ["LEVEL"]
                           [self.current_level]
                           ["LEARNBLOCKS"]
                           [self.current_lb]
                           ["DECONSTRUCTION"])

        if info.relationship == ("complete",):
            self.log_complete_deconstruction(info)

        elif info.relationship == ("Sigma", "T"):
            self.log_t_sigma_deconstruction(info)

        elif info.relationship == ("Sigma", "Z"):
            self.log_z_sigma_deconstruction(info)

        else:
            key["failed"] = {
                "inserted": [str(i.uid) for i in info.inserted],
                "deleted": [str(i.uid) for i in info.deleted]
            }

    def log_complete_deconstruction(self, info):
        """Store complete deconstruction log into journal.

        Args:
            info (DeconstructionInfo): Deconstruction log.

        """
        complete = (self.journal[self.current_n_block]
                                ["BLOCK"]
                                ["LEVEL"]
                                [self.current_level]
                                ["LEARNBLOCKS"]
                                [self.current_lb]
                                ["DECONSTRUCTION"]
                                ["COMPLETE"])

        complete.append({
            "state": str(info.state),
            "relative": str(info.relative.uid),
            "inserted": [str(i.uid) for i in info.inserted],
            "deleted": [str(i.uid) for i in info.deleted],
            "RECONSTRUCTED": []
        })

        pos = len(complete) - 1
        for recon_info in info.reconstructed_info:
            complete[pos]["RECONSTRUCTED"].append(
                {
                    "uid": str(recon_info.uid),
                    "subject": str(recon_info.subject),
                    "accuracy": str(recon_info.accuracy),
                    "reliability": str(recon_info.reliability)
                }
            )

    def log_t_sigma_deconstruction(self, info):
        """Store T Sigma deconstruction log into journal.

        Args:
            info (DeconstructionInfo): Deconstruction log.

        """
        t_sigma = (self.journal[self.current_n_block]
                              ["BLOCK"]
                              ["LEVEL"]
                              [self.current_level]
                              ["LEARNBLOCKS"]
                              [self.current_lb]
                              ["DECONSTRUCTION"]
                              ["T SIGMA"])

        t_sigma.append({
            "state": str(info.state),
            "relative": str(info.relative.uid),
            "LEARNBLOCK": {
                "rows": str(info.new_learnblock.n_rows()),
                "columns": str(info.new_learnblock.n_columns()),
                "min T": str(info.new_learnblock.min_timestamp()),
                "max T": str(info.new_learnblock.max_timestamp()),
                "origin": str(info.new_learnblock.origin)
            }
        })

    def log_z_sigma_deconstruction(self, info):
        """Store Z Sigma deconstruction log into journal.

        Args:
            info (DeconstructionInfo): Deconstruction log.

        """
        z_sigma = (self.journal[self.current_n_block]
                               ["BLOCK"]
                               ["LEVEL"]
                               [self.current_level]
                               ["LEARNBLOCKS"]
                               [self.current_lb]
                               ["DECONSTRUCTION"]
                               ["Z SIGMA"])

        z_sigma.append({
            "state": str(info.state),
            "relative": str(info.relative.uid),
            "inserted": [str(m.uid) for m in info.inserted if m is not None],
            "deleted": [str(m.uid) for m in info.deleted if m is not None]
        })



#  C O N S T R U C T I V I S T__ __
#   _____ ____   ____   /  |/  // /
#  / ___// __ \ / __ \ / /|_/ // /
# / /__ / /_/ // / / // /  / // /___
# \___/ \____//_/ /_//_/  /_//_____/
#  M A C H I N E   L E A R N I N G
#
#
# A Project by
# Thomas Schmid | UNIVERSITÄT LEIPZIG
# www.constructivist.ml
#
# Code Author: Dmitrij Denisenko
# Licence: MIT


"""Defines the wrapper for the ckmeans algorithms.

"""


def ckmeans(array, n_cluster):
    """Wrapper for the ckmeans algorithms.

    Args:
        array (list): One dimensional data set.
        n_cluster (int): Number of clusters.

    Returns:
        list:
            Clustered array.

    """
    return ckmeans_original(array, n_cluster)


# Copyright (c) 2016 Rocketrip.
def ssq(j, i, sum_x, sum_x_sq):
    if j > 0:
        muji = (sum_x[i] - sum_x[j-1]) / (i - j + 1)
        sji = sum_x_sq[i] - sum_x_sq[j-1] - (i - j + 1) * muji ** 2
    else:
        sji = sum_x_sq[i] - sum_x[i] ** 2 / (i+1)

    return 0 if sji < 0 else sji


def fill_row_k(imin, imax, k, S, J, sum_x, sum_x_sq, N):
    if imin > imax: return

    i = (imin+imax) // 2
    S[k][i] = S[k-1][i-1]
    J[k][i] = i

    jlow = k

    if imin > k:
        jlow = int(max(jlow, J[k][imin-1]))
    jlow = int(max(jlow, J[k-1][i]))

    jhigh = i-1
    if imax < N-1:
        jhigh = int(min(jhigh, J[k][imax+1]))

    for j in range(jhigh, jlow-1, -1):
        sji = ssq(j, i, sum_x, sum_x_sq)

        if sji + S[k-1][jlow-1] >= S[k][i]: break

        # Examine the lower bound of the cluster border
        # compute s(jlow, i)
        sjlowi = ssq(jlow, i, sum_x, sum_x_sq)

        SSQ_jlow = sjlowi + S[k-1][jlow-1]

        if SSQ_jlow < S[k][i]:
            S[k][i] = SSQ_jlow
            J[k][i] = jlow

        jlow += 1

        SSQ_j = sji + S[k-1][j-1]
        if SSQ_j < S[k][i]:
            S[k][i] = SSQ_j
            J[k][i] = j

    fill_row_k(imin, i-1, k, S, J, sum_x, sum_x_sq, N)
    fill_row_k(i+1, imax, k, S, J, sum_x, sum_x_sq, N)


def fill_dp_matrix(data, S, J, K, N):
    import numpy as np
    sum_x = np.zeros(N, dtype=np.float_)
    sum_x_sq = np.zeros(N, dtype=np.float_)

    # median. used to shift the values of x to improve numerical stability
    shift = data[N//2]

    for i in range(N):
        if i == 0:
            sum_x[0] = data[0] - shift
            sum_x_sq[0] = (data[0] - shift) ** 2
        else:
            sum_x[i] = sum_x[i-1] + data[i] - shift
            sum_x_sq[i] = sum_x_sq[i-1] + (data[i] - shift) ** 2

        S[0][i] = ssq(0, i, sum_x, sum_x_sq)
        J[0][i] = 0

    for k in range(1, K):
        if k < K-1:
            imin = max(1, k)
        else:
            imin = N-1

        fill_row_k(imin, N-1, k, S, J, sum_x, sum_x_sq, N)


def ckmeans_original(data, n_clusters):
    import numpy as np
    if n_clusters <= 0:
        raise ValueError("Cannot classify into 0 or less clusters")
    if n_clusters > len(data):
        raise ValueError("Cannot generate more classes than there are data values")

    # if there's only one value, return it; there's no sensible way to split
    # it. This means that len(ckmeans([data], 2)) may not == 2. Is that OK?
    unique = len(set(data))
    if unique == 1:
        return [data]

    data.sort()
    n = len(data)

    S = np.zeros((n_clusters, n), dtype=np.float_)

    J = np.zeros((n_clusters, n), dtype=np.uint64)

    fill_dp_matrix(data, S, J, n_clusters, n)

    clusters = []
    cluster_right = n-1

    for cluster in range(n_clusters-1, -1, -1):
        cluster_left = int(J[cluster][cluster_right])
        clusters.append(data[cluster_left:cluster_right+1])

        if cluster > 0:
            cluster_right = cluster_left - 1

    return list(reversed(clusters))



#  C O N S T R U C T I V I S T__ __
#   _____ ____   ____   /  |/  // /
#  / ___// __ \ / __ \ / /|_/ // /
# / /__ / /_/ // / / // /  / // /___
# \___/ \____//_/ /_//_/  /_//_____/
#  M A C H I N E   L E A R N I N G
#
#
# A Project by
# Thomas Schmid | UNIVERSITÄT LEIPZIG
# www.constructivist.ml
#
# Code Author: Dmitrij Denisenko
# Licence: MIT


"""Define the T-Sigma deconstruction.

"""


import sys
from dataclasses import dataclass
from queue import Empty
from multiprocessing import Process

from conML.shared.logger import DeconstructionInfo
from conML.domain.deconstruction.abstract import Abstract


__all__ = (
    "TSigmaDeconstructor",
    "Start",
    "Upgrade",
    "Apply",
    "Clean",
    "Exit"
)


@dataclass
class Start:
    """Serializable command, indicating to start the deconstruction. """
    level: int
    relatives: list
    pragmatic: object


@dataclass
class Upgrade:
    """Serializable command, indicating to upgrade the deconstructed models. """
    deletes: set


@dataclass
class Apply:
    """Serializable command, indicating to write down the changes. """
    pass


@dataclass
class Clean:
    """Serializable command, indicating to clean the temporary
    deconstruction changes.

    """
    pass


@dataclass
class Exit:
    """Serializable command, indicating to exit the deconstruction. """
    pass


class TSigmaDeconstructor(Abstract, Process):
    """Class performs the T Sigma deconstruction process.

    Args:
        build_new_learnblock (Callable):
            Function for building new PandasBlocks.
        krippendorf (Callable):
            Function for calculating krippendorff-alpha.
        knowledge (KnowledgeDatabase):
            Deconstruction changes writting to this database.
        settings (DeconstructionSettings):
            Controlls the behavior of the deconstruction.
        general_setitngs (GeneralSettings):
            Controlls the behavior of the general framework.
        inputs (Queue):
            Contains the inputs for the background deconstruction.
        outputs (Queue):
            Contained the outputs of the background deconstruction.
        upgrade_event (Event):
            Signals that the upgrades on the temporary changes are
            performed.
        apply_event (Event):
            Signals that the deconstructed objects are pushed to the
            output queue.
        clear_event (Event):
            Signals that the input and output queue is empty.

    """
    RELATION = ("Sigma", "T")

    def __init__(self, build_new_learnblock, krippendorf, knowledge, settings,
                 general_setitngs, inputs=None, outputs=None,
                 upgrade_event=None, apply_event=None, clear_event=None):
        super(Process, self).__init__(name="TSigmaBackgroundProcess")

        self.build_new_learnblock = build_new_learnblock
        self.krippendorf = krippendorf
        self.knowledge = knowledge
        self.settings = settings
        self.general_settings = general_setitngs
        self.temp_changes = []
        self.block = None
        self.relative_block = None

        # Used during parallelism
        self.inputs = inputs
        self.outputs = outputs
        self.upgrade_event = upgrade_event
        self.apply_event = apply_event
        self.clear_event = clear_event

    def run(self):
        """Performs T Sigma deconstruction in the background.

        """
        while True:
            command = self.inputs.get()

            if isinstance(command, Start):
                if not self.block:
                    self.block = command.pragmatic.trained_with(self.knowledge)

                for relative in command.relatives:
                    self.relative_block = relative.trained_with(self.knowledge)
                    log = self.log_info(relative, command.pragmatic)
                    self.deconstruct(command.level, relative, command.pragmatic,
                                     log)

            elif isinstance(command, Upgrade):
                self.upgrade(command.deletes)
                self.upgrade_event.set()

            elif isinstance(command, Apply):
                self.apply()
                self.apply_event.set()

            elif isinstance(command, Clean):
                self.clean()
                self.clear_event.set()

            elif isinstance(command, Exit):
                sys.exit(0)

    def run_serial(self, level, relatives, pragmatic, result_deque):
        """Performs the T Sigma deconstruction.

        Args:
            level (int):
                Level specification on which the deconstruction is performed.
            relatives (PragmaticMachineLearningModel):
                Relative of the model, which should be deconstructed.
            pragmatic (PragmaticMachineLearningModel):
                Model which should be deconstructed.
            result_deque (deque):
                Used as a container for all generated DeconstructionInfo.

        """
        self.block = pragmatic.trained_with(self.knowledge)

        for relative in relatives:
            self.relative_block = relative.trained_with(self.knowledge)
            log = self.log_info(relative, pragmatic)
            self.deconstruct(level, relative, pragmatic, log)

        for log, level, lb in self.temp_changes:
            ####################################################################
            self.knowledge.insert(pragmatic)
            log.inserted.append(pragmatic)
            ####################################################################

            result_deque.append((log, level, lb))

            if (log.state == DeconstructionInfo.State.DECONSTRUCTED
                    and self.settings.deconst_mode == "minimal"):
                self.temp_changes.clear()
                break

    def upgrade(self, deletions):
        """Remove invalid demporary deconstructions.

        Args:
            deletions (Iterable): List wich models, which have been
                removed during the complete deconstruction.

        """
        to_delete = []
        for log, level, learnblock in self.temp_changes:
            if set(deletions).intersection(set(learnblock.origin)):
                to_delete.append((log, level, learnblock))

        for pair in to_delete:
            self.temp_changes.remove(pair)

    def apply(self):
        """Push temporary deconstruction to the output queue.

        """
        for log, level, lb in self.temp_changes:
            self.outputs.put((log, level, lb))
            if (log.state == DeconstructionInfo.State.DECONSTRUCTED
                    and self.settings.deconst_mode == "minimal"):
                break

    def clean(self):
        """Clean the temporary changes from invalid changes.

        """
        self.temp_changes.clear()
        while not self.inputs.empty():
            try:
                self.inputs.get(block=False, timeout=None)
            except Empty:
                continue

        while not self.outputs.empty():
            try:
                self.outputs.get(block=False, timeout=None)
            except Empty:
                continue

    def deconstruct(self, level, relative, pragmatic, log):
        """Run the TSigma deconstruction.

        Args:
            level (int): Level on which deconstruction is performed.
            relative (PragmaticMachineLearningModel): Relative model.
            pragmatic (PragmaticMachineLearningModel): Model to deconstruct.
            log (DeconstructionInfo): Logging information.

        """
        if self.level_constraint(level):
            overlapping = self.identify_overlapping_block(self.block)
            if self.row_and_reliability_contraints(overlapping):
                block = self.prepare_higher_block(pragmatic, relative, overlapping)
                log.state = DeconstructionInfo.State.DECONSTRUCTED
                log.new_learnblock = block
                self.temp_changes.append((log, level + 1, block))

    def level_constraint(self, level):
        """Check if learnblock is not pushed over the maximal level.

        Args:
            level (int): Level where to push the learnblock.

        Returns:
            bool:
                True, if the maximal level is not reached else False.

        """
        return level < self.general_settings.highest_level

    def identify_overlapping_block(self, block):
        """Build learnblock on basis of overlapping timestamps.

        Args:
            block (PandasBlock): Learnblock on which the deconstructing model
                was trained.

        Returns:
            PandasBlock:
                Learnblock with overlapping time stamps.

        """
        overlapped_block = block.get_overlapping_on_timestamps(
            self.relative_block
        )
        return overlapped_block

    def row_and_reliability_contraints(self, block):
        """Check if all contrainsts are satisfied to push the learnblock
        to the next level.

        Args:
            block (PandasBlock): Learnblock on which the deconstructed model
                was trained.

        Returns:
            bool:
                True, if alle contraints are satisfied.

        """
        return self.row_constraint(block) and self.reliability_constraint(block)

    def row_constraint(self, block):
        """Check if the block has enough samples.

        Args:
            block (PandasBlock): Possible learnblock.

        Returns:
            bool:
                True if the block has enough samples to be considered a
                learnblock else False.

        """
        return len(block) >= self.settings.learn_block_minimum

    def reliability_constraint(self, block):
        """Check if the labels of the block satisfy the minimum
        inter-reliability.

        Args:
            block (PandasBlock): Possible learnblock.

        Returns:

        """
        predictions = [
            [int(i) for i in block.get_column("Z")],
            [int(i) for i in block.get_column("Z_other")]
        ]
        alpha = self.krippendorf(predictions)
        alpha_systematic = alpha < 0
        alpha_weak_reliability = (0 <= alpha < self.settings.min_reliability)
        return (self.settings.allow_weak_reliability and
                alpha_weak_reliability) or alpha_systematic

    def prepare_higher_block(self, pragmatic, relative, overlapping):
        """Build learnblock for the next level.

        Args:
            pragmatic (PragmaticMachineLearnignModel): Model to deconstruct.
            relative (PragmaticMachineLearningModel): Relative model.
            overlapping (PandasBlock): Learnblock that is valid to be pushed
                to the next level.

        Returns:
            PandasBlock

        """
        overblock = self.build_new_learnblock(
            values=list(zip(
                overlapping.get_column("Z"),
                overlapping.get_column("Z_other"),
                overlapping.get_column("T"),
                ("\"\"" for _ in range(overlapping.n_rows())),
                ("\"\"" for _ in range(overlapping.n_rows())))),
            columns=(pragmatic.uid, relative.uid, "T", "Sigma", "Z"),
            index=[i for i in range(overlapping.n_rows())],
            origin=(pragmatic.uid, relative.uid))
        return overblock



#  C O N S T R U C T I V I S T__ __
#   _____ ____   ____   /  |/  // /
#  / ___// __ \ / __ \ / /|_/ // /
# / /__ / /_/ // / / // /  / // /___
# \___/ \____//_/ /_//_/  /_//_____/
#  M A C H I N E   L E A R N I N G
#
#
# A Project by
# Thomas Schmid | UNIVERSITÄT LEIPZIG
# www.constructivist.ml
#
# Code Author: Dmitrij Denisenko
# Licence: MIT


from conML.domain.preprocessing import *
from conML.domain.construction import *
from conML.domain.reconstruction import *
from conML.domain.deconstruction.deconstruction import *
from conML.domain.reconstruction.selection import *


__all__ = (
    "Constructor",
    "Reconstructor",
    "Deconstructor",
    "FeatureSelector",
    "ClusterFinder"
)



#  C O N S T R U C T I V I S T__ __
#   _____ ____   ____   /  |/  // /
#  / ___// __ \ / __ \ / /|_/ // /
# / /__ / /_/ // / / // /  / // /___
# \___/ \____//_/ /_//_/  /_//_____/
#  M A C H I N E   L E A R N I N G
#
#
# A Project by
# Thomas Schmid | UNIVERSITÄT LEIPZIG
# www.constructivist.ml
#
# Code Author: Dmitrij Denisenko
# Licence: MIT


from conML.ports.krippendorff_adapter import krippendorff_alpha
from conML.ports.ckmeans_adapter import ckmeans
from conML.ports.source_adapter import (
    convert_df_to_block,
    build_block_from_rows,
    build_new_learnblock
)
from conML.ports.ml_adapter import (
    ConstructionClusteringMLModel,
    ReconstructionConceptualMLModel,
    FilterMethod,
    EmbeddedMethod,
    KernelDensityEstimator
)

__all__ = (
    "ConstructionClusteringMLModel",
    "ReconstructionConceptualMLModel",
    "FilterMethod",
    "EmbeddedMethod",
    "KernelDensityEstimator",
    "convert_df_to_block",
    "build_new_learnblock",
    "build_block_from_rows",
    "ckmeans",
    "krippendorff_alpha"
)



"""Defines the public interface to be used.

Every new functionality that is added to the system and capable of being
influenced by the user should have an interface.
The main functionality provided by this module is:
- Load settings file from a specific path.
- Returning the currently used settings.
- Create Construction object and initialize it with user specified arguments.
- Create FeatureSelector object and initialize it with user specified arguments.
- Create Reconstruction object and initialize it with user specified arguments.
- Create Deconstruction object and initialize it with user specified arguments.
- Start the learning process depending on user specified Construction,
Reconstruction, Deconstruction object.
- Load KnowledgeDatabase from a specific destination.

"""


import os
import pickle

from conML import ports
from conML.shared import settings
from conML.shared import request
from conML.usecases import command, query


__all__ = (
    "load_settings",
    "get_settings",
    "construction",
    "reconstruction",
    "reconstruction",
    "deconstruction",
    "knowledge_searcher",
    "load_knowledge",
    "feature_selection",
)


def load_settings(path):
    """Load settings from a source path and use them across the whole program.

    Args:
        path (str): Source path to the settings files.

    Returns:
        None

    """
    assert isinstance(path, str)
    assert os.path.exists(path)
    settings.read_settings(path)


def get_settings():
    """Return currently used settings.

    Returns:
        Settings:
            Currently used settings object.

    """
    return settings.Settings


def load_knowledge(path):
    """Load knowledge database from a source path.

    Args:
        path (str): Source path to the knowledge database.

    Returns:
        KnowledgeDatabase:
            Collected knowledge representation from learning process saved
            into database.

    """
    assert isinstance(path, str)
    assert os.path.exists(path)
    with open(path, "rb") as file:
        return pickle.load(file)


def construction(mode, algorithms):
    """Create construction object depending on user provided mode and
    algorithms. Algorithms should be a list of sklearn objects, which
    must define unsupervised machine learning methods.

    Note:
        Only "conceptual" mode is currently supported.

    Args:
        mode (str): Either "conceptual" or "procedural".
        algorithms (list): List with unsupervised learning algorithms from
            sklearn library.

    Returns:
        Constructor:
            Object defines construction part of learning process in
            constructivism machine learning.

    """
    assert mode == "conceptual"
    interface = ports.ConstructionClusteringMLModel
    setting = settings.specific_settings_factory("construction")
    req = request.ConstructionRequest(setting, algorithms, interface, mode)
    usecase = command.ConstructionUsecase()
    return usecase.execute(req)


def feature_selection(*, filter_method, embedded_method):
    """Create feature selection object depending on user provided filter
    and embedded methods.

    Args:
        filter_method: Filter method for feature reduction/selection from
            sklearn library.
        embedded_method: Embedded method for feature reduction/selection from
            sklearn library.

    Returns:
        FeatureSelector:
            Object define feature selection part within reconstruction process
            in constructivism machine learning.

    """
    assert filter_method is not None
    assert embedded_method is not None
    filter_method = ports.FilterMethod(filter_method)
    embedded_method = ports.EmbeddedMethod(embedded_method)
    setting = settings.specific_settings_factory("feature_selection")
    req = request.FeatureSelectionRequest(filter_method, embedded_method, setting)
    usecase = command.FeatureSelectionUsecase()
    return usecase.execute(req)


def reconstruction(mode, algorithms):
    """Create reconstruction object depending on user provided mode and
    algorithms. Algorithms should be a list of sklearn objects, which
    must define supervised machine learning methods.

    Note:
        Only "conceptual" mode is currently supported.

    Args:
        mode (str): Either "conceptual" or "procedural".
        algorithms (list): List with supervised learning algorithms from
            sklearn library.

    Returns:
        Reconstructor:
            Object defines reconstruction part of learning process in
            constructivism machine learning.

    """
    assert mode == "conceptual"
    interface = ports.ReconstructionConceptualMLModel
    setting = settings.specific_settings_factory("reconstruction")
    req = request.ReconstructionRequest(setting, algorithms, interface,
                                        mode, ports.krippendorff_alpha)
    usecase = command.ReconstructionUsecase()
    return usecase.execute(req)


def deconstruction(reconstructor):
    """Create deconstruction object depending on user provided mode.

    Note:
        Only "conceptual" mode is currently supported.

    Args:
        reconstructor (Reconstructor): Coherent instantiated Reconstructor
            object.

    Returns:
        Deconstructor:
            Object defines deconstruction part of learning process in
            constructivism machine learning.

    """
    general_settings = settings.specific_settings_factory("general")
    deconstruction_settings = settings.specific_settings_factory("deconstruction")
    block_processing_settings = settings.specific_settings_factory("block_processing")
    density_estimator = ports.KernelDensityEstimator()
    req = request.DeconstructionRequest(general_settings,
                                        deconstruction_settings,
                                        density_estimator,
                                        ports.ckmeans,
                                        block_processing_settings,
                                        ports.build_block_from_rows,
                                        ports.build_new_learnblock,
                                        reconstructor)
    usecase = command.DeconstructionUsecase()
    return usecase.execute(req)


def knowledge_searcher(constructor,
                       selector,
                       reconstructor,
                       deconstructor,
                       knowledge_database=None,
                       *,
                       stdout=False,
                       n_procs=1):
    """Start the learning process depending on user provided arguments.
    All components must have coherent modes. It is illegal to start the
    learning process with conceptual and procedural components. The user
    can follow the learning process on the stdout by flagging stdout parameter
    to True. By flagging debug to True the learning process will be logged
    in detail to a file. Further the user can speed up the process by
    utilizing many cpu cores by setting parallel to True.

    Args:
        constructor (Constructor): Coherent instantiated Constructor object.
        selector (FeatureSelector): Instantiated FeatureSelector object.
        reconstructor (Reconstructor): Coherent instantiated Reconstructor
            object.
        deconstructor (Deconstructor): Coherent instantiated Deconstructor
            object.
        knowledge_database (KnowledgeDatabase): ...
        stdout (bool): Indicator if the learning process should be logged
            to stdout.
        n_procs (int): Number of processes to utilize.

    Returns:
        KnowledgeDatabase:
            Modified Knowledge representation after the learning process.

    """
    general_settings = settings.specific_settings_factory("general")
    block_processing_settings = settings.specific_settings_factory("block_processing")
    density_estimator = ports.KernelDensityEstimator()
    req = request.KnowledgeSearchingRequest(general_settings,
                                            block_processing_settings,
                                            density_estimator,
                                            ports.convert_df_to_block,
                                            ports.build_block_from_rows,
                                            ports.ckmeans,
                                            constructor,
                                            selector,
                                            reconstructor,
                                            deconstructor,
                                            knowledge_database,
                                            stdout, n_procs)

    usecase = query.KnowledgeSearcherUsecase()
    return usecase.execute(req)
